"protocol for ark"

from __future__ import annotations

from enum import Enum
from typing import (
    Any,
    Coroutine,
    Dict,
    List,
    Optional,
    Protocol,
    Tuple,
    Union,
)

import volcenginesdkarkruntime.types.chat.chat_completion_chunk as completion_chunk
from pydantic import BaseModel, Field, field_validator, model_validator
from typing_extensions import Annotated, Literal
from volcenginesdkarkruntime.types.chat.chat_completion import ChatCompletion, Choice
from volcenginesdkarkruntime.types.chat.chat_completion_stream_options_param import (
    ChatCompletionStreamOptionsParam,
)
from volcenginesdkarkruntime.types.completion_usage import CompletionUsage

from ark.core.idl.common_protocol import (
    ActionDetails,
    ActionUsage,
    BotUsage,
    ChatUsage,
    Error,
    LabelLogprobosValue,
    ModelDesc,
    Reference,
    Request,
    Response,
)
from ark.core.utils.errorsv3 import InvalidParameter, MissingParameter


class CallableFunction(Protocol):
    def __call__(
        self, request: Dict[str, Any], **kwargs: Any
    ) -> Coroutine[
        Any,
        Any,
        Union[Union[str, BaseModel], Tuple[Union[str, BaseModel], ActionDetails]],
    ]:
        ...


EndpointId = str


class FunctionCallMode(str, Enum):
    SEQUENTIAL = "sequential"
    """
    The function will be called sequentially, one after the other.
    """
    PARALLEL = "parallel"
    """
    The function will be called in parallel, and the results will be combined at the
    end.
    """
    ONCE = "ONCE"
    """
    The function will be called once and returns the result.
    """


class FunctionDefinition(BaseModel):
    name: str
    """The name of the function to be called.

    Must be a-z, A-Z, 0-9, or contain underscores and dashes, with a maximum length
    of 64.
    """

    description: str
    """
    A description of what the function does, used by the model to choose when and
    how to call the function.
    """

    parameters: Dict[str, Any]
    """The parameters the functions accepts, described as a JSON Schema object.

    See the
    [JSON Schema reference](https://json-schema.org/understanding-json-schema/) for
    documentation about the format.

    Omitting `parameters` defines a function with an empty parameter list.
    """


class ChatCompletionTool(BaseModel):
    function: FunctionDefinition

    type: Literal["function"]
    """The type of the tool. Currently, only `function` is supported."""


class ArkChatParameters(BaseModel):
    frequency_penalty: Optional[float] = None
    """
    Number between -2.0 and 2.0. Positive values penalize new tokens based on their
    existing frequency in the text so far, decreasing the model's likelihood to
    repeat the same line verbatim.
    """
    logit_bias: Optional[Dict[str, int]] = None
    """
    Modify the likelihood of specified tokens appearing in the completion.
    """
    logprobs: Optional[bool] = None
    max_tokens: Optional[int] = None
    presence_penalty: Optional[float] = None
    stop: Optional[Union[Optional[str], List[str]]] = None
    stream_options: Optional[ChatCompletionStreamOptionsParam] = None
    temperature: Optional[float] = None
    tools: Optional[List[ChatCompletionTool]] = None
    top_logprobs: Optional[int] = None
    top_p: Optional[float] = None

    n: Optional[int] = Field(default=1, ge=1, le=5)
    """
    How many chat completion choices to generate for each input message. 
    Note that you will be charged based on the number of generated tokens 
    across all of the choices. 

    Keep n as 1 to minimize costs.
    """

    def merge_from(
        self, other: Union[Dict[str, Any], ArkChatParameters]
    ) -> ArkChatParameters:
        other_dict = other if isinstance(other, dict) else other.dict()
        merged_dict = self.dict()

        for key in merged_dict.keys():
            value = other_dict.get(key, None)
            if value is not None:
                merged_dict[key] = value

        return self.__class__(**merged_dict)

    def merge_to(
        self, other: Union[Dict[str, Any], ArkChatParameters]
    ) -> ArkChatParameters:
        self_dict = self.dict()
        merged_dict = other if isinstance(other, dict) else other.dict()

        for key in self_dict.keys():
            value = self_dict.get(key, None)
            if value is not None:
                merged_dict[key] = value

        return self.__class__(**merged_dict)


class Function(BaseModel):
    arguments: str
    """
    The arguments to call the function with, as generated by the model in JSON
    format. Note that the model does not always generate valid JSON, and may
    hallucinate parameters not defined by your function schema. Validate the
    arguments in your code before calling your function.
    """

    name: str
    """The name of the function to call."""


class ChatCompletionMessageToolCallParam(BaseModel):
    id: str
    """The ID of the tool call."""

    function: Function
    """The function that the model called."""

    type: Literal["function"]
    """The type of the tool. Currently, only `function` is supported."""


class ChatCompletionMessageImageUrlPartImageUrl(BaseModel):
    url: str
    detail: Optional[str] = None


class ChatCompletionMessageImageUrlPart(BaseModel):
    type: Literal["image_url"]
    image_url: ChatCompletionMessageImageUrlPartImageUrl


class ChatCompletionMessageTextPart(BaseModel):
    type: Literal["text"]
    text: str


ChatCompletionMessagePart = Annotated[
    Union[ChatCompletionMessageImageUrlPart, ChatCompletionMessageTextPart],
    Field(discriminator="type"),
]


class ArkMessage(BaseModel):
    role: Literal["user", "system", "assistant", "tool"]
    content: Union[str, List[ChatCompletionMessagePart]]
    name: Optional[str] = None
    tool_call_id: Optional[str] = None
    tool_calls: Optional[List[ChatCompletionMessageToolCallParam]] = None

    @classmethod
    @model_validator(mode="before")
    def validate_content(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        role, content = v.get("role"), v.get("content")
        if not isinstance(content, str) and role != "user":
            raise InvalidParameter(
                parameter="content",
                cause=f"content must be type of str when role is {role}",
            )

        tool_call_id = v.get("tool_call_id")
        if tool_call_id is not None and role != "tool":
            raise InvalidParameter(
                parameter="tool_call_id",
                cause=f"tool_call_id must be None when role is {role}",
            )

        tool_calls = v.get("tool_calls")
        if tool_calls is not None and role != "assistant":
            raise InvalidParameter(
                parameter="tool_calls",
                cause=f"tool_calls must be None when role is {role}",
            )
        return v


class ArkChatRequest(Request):
    messages: List[ArkMessage]
    """
    chat messages 
    """
    model: str
    """
    bot id
    """
    frequency_penalty: Optional[float] = None
    """
    Number between -2.0 and 2.0. Positive values penalize new tokens based on their
    existing frequency in the text so far, decreasing the model's likelihood to
    repeat the same line verbatim.
    """
    logit_bias: Optional[Dict[str, int]] = None
    """
    Modify the likelihood of specified tokens appearing in the completion.
    """
    logprobs: Optional[bool] = None
    max_tokens: Optional[int] = None
    presence_penalty: Optional[float] = None
    stop: Optional[Union[Optional[str], List[str]]] = None
    stream: bool = False
    stream_options: Optional[ChatCompletionStreamOptionsParam] = None
    temperature: Optional[float] = None
    tools: Optional[List[ChatCompletionTool]] = None
    top_logprobs: Optional[int] = None
    top_p: Optional[float] = None
    user: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = Field(default_factory=dict)

    n: Optional[int] = Field(default=1, ge=1, le=5)
    """
    How many chat completion choices to generate for each input message. 
    Note that you will be charged based on the number of generated tokens 
    across all of the choices. 

    Keep n as 1 to minimize costs.
    """

    def get_user_info_extra(self) -> Optional[UserInfoExtra]:
        if not self.metadata or not self.metadata.get("user_info"):
            return None
        user_info = self.metadata.get("user_info", "")
        if isinstance(user_info, str):
            return UserInfoExtra.model_validate_json(user_info)
        else:
            return UserInfoExtra.model_validate(user_info)

    def is_emit_intention_signal_extra(self) -> bool:
        if not self.metadata:
            return False
        intention_signal = self.metadata.get("emit_intention_signal_extra", "false")
        if isinstance(intention_signal, bool):
            return intention_signal
        else:
            return intention_signal == "true"

    def get_chat_request(
        self, extra_body: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        dumped_dict = self.model_dump(exclude_unset=True, exclude_none=True)

        extra_body = extra_body or {}
        if "metadata" in dumped_dict:
            extra_body["metadata"] = dumped_dict.pop("metadata")
        dumped_dict["extra_body"] = extra_body
        return dumped_dict

    @classmethod
    @field_validator("messages")
    def validate_messages(
        cls, v: Optional[List[ArkMessage]]
    ) -> Optional[List[ArkMessage]]:
        if not v:
            raise MissingParameter("messages")

        for msg in v:
            if not msg.role:
                raise MissingParameter("role")
            if msg.content is None:
                msg.content = ""

            if msg.role == "tool" and not msg.tool_call_id:
                raise MissingParameter("tool_call_id")

        return v


class ArkChatResponse(Response):
    id: str
    """A unique identifier for the chat completion."""

    choices: List[Choice]
    """A list of chat completion choices.

    Can be more than one if `n` is greater than 1.
    """

    created: int
    """The Unix timestamp (in seconds) of when the chat completion was created."""

    model: str
    """The model used for the chat completion."""

    object: Literal["chat.completion"]
    """The object type, which is always `chat.completion`."""

    usage: Optional[CompletionUsage] = None
    """Usage statistics for the completion request."""

    bot_usage: Optional[BotUsage] = None

    metadata: Optional[Dict[str, Any]] = None

    references: Optional[List[Reference]] = None

    @staticmethod
    def merge(
        responses: List[Union[ArkChatResponse, ChatCompletion]],
    ) -> ArkChatResponse:
        assert len(responses) > 0, "empty responses"

        """
        to ensure the `merged` have attributes `usage`
        handle the responses in reversed order
        """
        merged = ArkChatResponse(**responses[-1].__dict__)
        for resp in reversed(responses[:-1]):
            for i, j in zip(merged.choices, resp.choices):
                if isinstance(i.message.content, str) and isinstance(
                    j.message.content, str
                ):
                    i.message.content = j.message.content + i.message.content
                elif isinstance(i.message.content, list) and isinstance(
                    j.message.content, list
                ):
                    i.message.content = j.message.content + i.message.content
                else:
                    raise TypeError("no supported merge type")

            if merged.usage and resp.usage:
                merged.usage.prompt_tokens += resp.usage.prompt_tokens
                merged.usage.completion_tokens += resp.usage.completion_tokens
                merged.usage.total_tokens += resp.usage.total_tokens
            if merged.bot_usage and resp.bot_usage:
                merged.bot_usage += resp.bot_usage
        return merged

    def transform_bot_usage(self) -> None:
        """
        used for transforming bot usage without actions
        """

        if not self.usage or self.bot_usage:
            return

        self.bot_usage = BotUsage(
            model_usage=[
                ChatUsage(
                    prompt_tokens=self.usage.prompt_tokens,
                    completion_tokens=self.usage.completion_tokens,
                    total_tokens=self.usage.total_tokens,
                )
            ]
        )
        return

    def merge_usages(
        self, others: Union[CompletionUsage, List[CompletionUsage]]
    ) -> CompletionUsage:
        if not others:
            return self.usage

        if not isinstance(others, list):
            others = [others]

        total_usage = CompletionUsage(
            prompt_tokens=self.usage.prompt_tokens if self.usage else 0,
            completion_tokens=self.usage.completion_tokens if self.usage else 0,
            total_tokens=self.usage.total_tokens if self.usage else 0,
        )
        for usage in others:
            total_usage.prompt_tokens += usage.prompt_tokens
            total_usage.completion_tokens += usage.completion_tokens
            total_usage.total_tokens += usage.total_tokens

        self.usage = total_usage
        return total_usage

    def merge_bot_usage(self, others: Union[Any, List[Any]]) -> Optional[BotUsage]:
        if not others:
            return self.bot_usage

        if not self.bot_usage:
            self.bot_usage = BotUsage()

        if not isinstance(others, list):
            others = [others]

        for other in others:
            if isinstance(other, ActionDetails):
                if self.bot_usage.action_details is None:
                    self.bot_usage.action_details = []
                self.bot_usage.action_details.append(other)

                if self.bot_usage.action_usage is None:
                    self.bot_usage.action_usage = []
                self.bot_usage.action_usage.append(
                    ActionUsage(action_name=other.name, count=other.count)
                )
            elif isinstance(other, (ChatUsage, CompletionUsage)):
                if self.bot_usage.model_usage is None:
                    self.bot_usage.model_usage = []
                self.bot_usage.model_usage.append(other)
            elif isinstance(other, BotUsage):
                self.bot_usage += other
            elif (
                isinstance(other, tuple)
                and len(other) == 2
                and isinstance(other[1], ActionDetails)
            ):
                response, action_detail = other
                # deprecated
                if isinstance(response, ArkSearchResponse) and response.search_usage:
                    if self.bot_usage.action_usage is None:
                        self.bot_usage.action_usage = []
                    self.bot_usage.action_usage.extend(response.search_usage)

                self.merge_bot_usage(action_detail)
        return self.bot_usage


class ArkChatCompletionChunk(Response):
    id: str
    """A unique identifier for the chat completion. Each chunk has the same ID."""

    choices: List[completion_chunk.Choice]
    """A list of chat completion choices.

    Can be more than one if `n` is greater than 1.
    """

    created: int
    """The Unix timestamp (in seconds) of when the chat completion was created.

    Each chunk has the same timestamp.
    """

    model: str
    """The model to generate the completion."""

    object: Literal["chat.completion.chunk"]
    """The object type, which is always `chat.completion.chunk`."""

    usage: Optional[CompletionUsage] = None
    """
    An optional field that will only be present when you set
    `stream_options: {"include_usage": true}` in your request. When present, it
    contains a null value except for the last chunk which contains the token usage
    statistics for the entire request.
    """

    bot_usage: Optional[BotUsage] = None

    metadata: Optional[Dict[str, Any]] = None

    references: Optional[List[Reference]] = None

    def transform_bot_usage(self) -> None:
        """
        used for transforming bot usage without actions
        """

        if not self.usage or (
            self.bot_usage
            and self.bot_usage.model_usage
            and len(self.bot_usage.model_usage) > 0
            and self.bot_usage.model_usage[0].total_tokens
        ):
            return
        if self.bot_usage is None:
            self.bot_usage = BotUsage()
        self.bot_usage.model_usage = [
            ChatUsage(
                prompt_tokens=self.usage.prompt_tokens,
                completion_tokens=self.usage.completion_tokens,
                total_tokens=self.usage.total_tokens,
            )
        ]
        return

    @staticmethod
    def merge(
        responses: List[
            Union[ArkChatCompletionChunk, completion_chunk.ChatCompletionChunk]
        ],
    ) -> Optional[ArkChatCompletionChunk]:
        if len(responses) == 0:
            return None

        """
        to ensure the `merged` have attributes `usage`
        handle the responses in reversed order
        """
        merged = ArkChatCompletionChunk(**responses[-1].__dict__)
        if not merged.bot_usage:
            merged.bot_usage = BotUsage(
                model_usage=[], action_usage=[], action_details=[]
            )
        for resp in reversed(responses[:-1]):
            for i, j in zip(merged.choices, resp.choices):
                if isinstance(i.delta.content, str) and isinstance(
                    j.delta.content, str
                ):
                    i.delta.content = j.delta.content + i.delta.content
                elif isinstance(i.delta.content, list) and isinstance(
                    j.delta.content, list
                ):
                    i.delta.content = j.delta.content + i.delta.content
                else:
                    raise TypeError("no supported merge type")

            if merged.usage and resp.usage:
                merged.usage.prompt_tokens += resp.usage.prompt_tokens
                merged.usage.completion_tokens += resp.usage.completion_tokens
                merged.usage.total_tokens += resp.usage.total_tokens
            if (
                merged.bot_usage
                and isinstance(resp, ArkChatCompletionChunk)
                and resp.bot_usage
            ):
                merged.bot_usage += resp.bot_usage
        return merged

    def merge_usages(
        self, others: Union[CompletionUsage, List[CompletionUsage]]
    ) -> CompletionUsage:
        if not others:
            return self.usage

        if not isinstance(others, list):
            others = [others]

        total_usage = CompletionUsage(
            prompt_tokens=self.usage.prompt_tokens if self.usage else 0,
            completion_tokens=self.usage.completion_tokens if self.usage else 0,
            total_tokens=self.usage.total_tokens if self.usage else 0,
        )
        for usage in others:
            total_usage.prompt_tokens += usage.prompt_tokens
            total_usage.completion_tokens += usage.completion_tokens
            total_usage.total_tokens += usage.total_tokens

        self.usage = total_usage
        return total_usage

    def merge_bot_usage(self, others: Union[Any, List[Any]]) -> Optional[BotUsage]:
        if not others:
            return self.bot_usage

        if not self.bot_usage:
            self.bot_usage = BotUsage()

        if not isinstance(others, list):
            others = [others]

        for other in others:
            if isinstance(other, ActionDetails):
                if self.bot_usage.action_details is None:
                    self.bot_usage.action_details = []
                self.bot_usage.action_details.append(other)

                if self.bot_usage.action_usage is None:
                    self.bot_usage.action_usage = []
                self.bot_usage.action_usage.append(
                    ActionUsage(action_name=other.name, count=other.count)
                )
            elif isinstance(other, (ChatUsage, CompletionUsage)):
                if self.bot_usage.model_usage is None:
                    self.bot_usage.model_usage = []
                self.bot_usage.model_usage.append(other)
            elif isinstance(other, BotUsage):
                self.bot_usage += other
            elif (
                isinstance(other, tuple)
                and len(other) == 2
                and isinstance(other[1], ActionDetails)
            ):
                response, action_detail = other
                # deprecated
                if isinstance(response, ArkSearchResponse) and response.search_usage:
                    if self.bot_usage.action_usage is None:
                        self.bot_usage.action_usage = []
                    self.bot_usage.action_usage.extend(response.search_usage)

                self.merge_bot_usage(action_detail)
        return self.bot_usage


class ArkClassificationRequest(Request):
    query: Optional[str] = None
    labels: List[str]


class ArkClassificationResponse(Response):
    label: Optional[str] = None
    label_logprobos: Optional[Dict[str, LabelLogprobosValue]] = None
    usage: Optional[ChatUsage] = None
    error: Optional[Error] = None


class ArkTokenizeRequest(Request):
    text: Optional[str] = None


class ArkTokenizeResponse(Response):
    total_tokens: Optional[int] = None
    tokens: Optional[List[str]] = None
    error: Optional[Error] = None


class ArkEmbeddingsRequest(Request):
    input: Union[str, List[str]]
    encoding_format: Literal["float", "base64"] = "float"
    user: Optional[str] = None


class UserInfoExtra(BaseModel):
    city: str = ""
    district: str = ""


class ArkActionMeta(Request):
    action_name: str = ""
    model: Optional[str] = None
    function_definition: Optional[FunctionDefinition] = None


class ArkToolMeta(Request):
    action_name: str = ""
    tool_name: str = ""
    model: Optional[str] = None


SourceType = Literal[
    "weather",
    "douyin_baike",
    "toutiao_article",
    "douyin_short_video",
    "search_engine",
    "xigua_feed_video",
    "toutiao_short_content",
]

SOURCE_TAG = {
    "search_engine": "搜索引擎",
    "weather": "墨迹天气",
    "toutiao_article": "头条图文",
    "douyin_short_video": "抖音视频",
    "douyin_baike": "抖音百科",
    "xigua_feed_video": "西瓜视频",
    "toutiao_short_content": "微头条",
}

SearchAction = Literal[
    "WebBrowsing", "WebBrowsingPlus", "WebBrowsingPro", "content_plugin"
]


class ArkSearchSourceType(str, Enum):
    DEFAULT_SEARCH = "search_engine"
    WEATHER = "weather"
    TOUTIAO_ARTICLE = "toutiao_article"
    DOUYIN_SHORT_VIDEO = "douyin_short_video"
    DOUYIN_BAIKE = "douyin_baike"
    XIGUA_FEED_VIDEO = "xigua_feed_video"
    TOUTIAO_SHORT_CONTENT = "toutiao_short_content"


class ArkSearchRequest(Request):
    question: str = ""
    action_name: Literal["WebBrowsingPlus", "WebBrowsingPro", "content_plugin"] = Field(
        default="WebBrowsingPlus"
    )
    source_type: Optional[List[SourceType]] = None
    summary_top_k: int = 10
    utm_source: Optional[str] = None
    keywords: Optional[List[str]] = None
    user_info: Optional[UserInfoExtra] = None
    enhance_param: Optional[Dict[str, Any]] = None


class ArkKnowledgeBaseRequest(Request):
    collection_name: str
    project: str = "default"
    question: str = ""
    retrieve_count: int = 5
    keywords: Optional[List[str]] = None
    rerank_retrieve_count: Optional[int] = None
    query_param: Optional[Dict[str, Any]] = None
    dense_weight: Optional[float] = 0.5
    rerank_switch: bool = False
    show_reference: bool = True


class ArkSearchResponse(Response):
    search_results: List[str] = Field(default_factory=list)
    references: List[Reference] = Field(default_factory=list)
    search_usage: Optional[List[ActionUsage]] = None


class ArkKnowledgeBaseResponse(Response):
    texts: List[str] = Field(default_factory=list)
    references: List[Reference] = Field(default_factory=list)


class ArkSearchIntentionMetadata(ArkActionMeta):
    action_name: SearchAction = Field(default="WebBrowsing")
    result_mapping: Optional[Dict[str, bool]] = None
    keywords: Optional[List[str]] = None


class ArkSearchSummaryMetadata(ArkActionMeta):
    action_name: SearchAction = Field(default="WebBrowsing")
    keywords: Optional[List[str]] = None
    summary_top_k: int = 10
    summary_max_total_len: Optional[int] = None

    biz_id: Literal["seed", "volc"] = "seed"
    extra_params: Optional[Dict[str, Any]] = None
    ab_params: Optional[Dict[str, Any]] = None

    browsing_doc_list: Optional[str] = Field(default="")
    summary_with_system_prompt: Optional[bool] = Field(default=False)
    user_info: Optional[UserInfoExtra] = None
    site: Optional[str] = Field(default="")
    allow_host: Optional[List[str]] = None
    block_host: Optional[List[str]] = None
    publish_time: Dict[str, str] = Field(default_factory=dict)


class ArkCalculatorRequest(Request):
    input: str = ""
    query: Optional[str] = None


class ArkCalculatorResponse(Response):
    result: str = ""
    message: Optional[str] = None


class ArkWebData(BaseModel):
    url: str = ""
    title: str = ""
    content: str = ""
    error_message: Optional[str] = None
    error_code: Optional[str] = None


class ArkLinkReaderRequest(Request):
    url_list: List[str] = Field(default_factory=list)


class ArkLinkReaderResponse(Response):
    ark_web_data_list: List[ArkWebData] = Field(default_factory=list)


class ArkCharacterConfig(BaseModel):
    # character
    name: str = Field(
        default="",
        description="character name, should be unique in single group chat config",
    )
    system_prompt: str = Field(default="", description="character description")
    model_desc: Optional[ModelDesc] = None


class ArkGroupChatConfig(BaseModel):
    # group chat
    description: Optional[str] = Field(
        default="", description="group chat scene description"
    )
    characters: List[ArkCharacterConfig] = Field(
        default=[], description="group chat character configs"
    )
    user_name: Optional[str] = Field(
        default="", description="user character name, optional"
    )


class ArkToolRequest(ArkToolMeta):
    dry_run: Optional[bool] = False
    timeout: Optional[int] = 60
    parameters: Optional[Dict[str, Any]] = None


class ArkToolResponse(Response):
    status_code: Optional[int] = None
    data: Optional[Any] = None



from typing import Dict, List, Optional

from pydantic import Field

from ark.core.config import ActionConfig
from ark.core.idl.ark_protocol import SourceType


class SearchConfig(ActionConfig):
    # browsing
    result_mapping: Optional[Dict[str, bool]] = None
    keywords: Optional[List[str]] = None
    source_type: Optional[List[SourceType]] = None
    utm_source: Optional[str] = Field(default="")
    summary_top_k: int = 10



from ark.core.config.action_config import ActionConfig, AtomActionConfig
from ark.core.config.config_center import (
    CONFIG_CENTER_DIR,
    CONFIG_CENTER_NAME,
    ConfEnv,
    ConfigCenter,
    PipelineConfig,
    Settings,
)
from ark.core.config.endpoint_config import (
    RAG_SOURCE,
    RAG_TAG,
    SOURCE_PRIPORITY,
    EndpointConfig,
    PhaseMode,
    Prompt,
    default_endpoint_config,
)

__all__ = [
    "EndpointConfig",
    "PhaseMode",
    "default_endpoint_config",
    "SOURCE_PRIPORITY",
    "RAG_SOURCE",
    "RAG_TAG",
    "ActionConfig",
    "AtomActionConfig",
    "ConfigCenter",
    "Settings",
    "PipelineConfig",
    "CONFIG_CENTER_DIR",
    "CONFIG_CENTER_NAME",
    "ConfEnv",
    "Prompt",
]



import json
import logging
from typing import List

from ark.component.reader._reader import Reader
from ark.core.rag import KnowledgeChunk, KnowledgeDoc
from ark.core.task import task

logging.basicConfig(
    level=logging.INFO, format="[%(asctime)s][%(levelname)s] %(message)s"
)
LOGGER = logging.getLogger(__name__)


class FeishuReader(Reader):
    """
    FeishuReader for knowledge base
    """

    def _rechunk(
        self,
        lark_chunks: List[str],
        full_text_chunk: KnowledgeChunk,
        chunk_merge_separator: str = "\n",
        need_encode_title: bool = False,
        custom_prefix: str = "",
    ) -> List[KnowledgeChunk]:
        encode_text_prefix = custom_prefix
        title = full_text_chunk.scalars["title"]
        doc_id = full_text_chunk.scalars["doc_id"]
        tos_url = full_text_chunk.scalars["tos_url"]
        file_name = full_text_chunk.scalars["file_name"]
        full_text_len = full_text_chunk.scalars["full_text_len"]
        primary_key = self.config.primary_key.get("name", "")
        vector_key = self.config.vector.get("name", "")

        if need_encode_title:
            encode_text_prefix = f"{custom_prefix}标题:{title}\n"
        new_chunk_list = []
        new_encode_text = encode_text_prefix
        new_origin_text = ""
        new_chunk_page_num = []
        new_chunk_id = 0

        chunk_type = "raw_chunk"
        for chunk_id, chunk in enumerate(lark_chunks):
            chunk_root = json.loads(chunk)
            if chunk_root["type"] not in [
                "section-text",
                "section-title",
                "footnote",
            ]:
                continue
                # title：全文大标题, section-title：章节标题, section-text：章节内容
                # image：图片, table：表格, header：页眉, footer：页脚
                # footnote：脚注, caption：图/表描述, toc：目录, others：其他
            cur_text = chunk_root.get("text")
            cur_encode_text_len = len(new_encode_text + cur_text)

            if (
                new_encode_text == encode_text_prefix
                or cur_encode_text_len <= self.split_size
            ):
                # Case1: new_encode_text == chunk_prefix, len(cur_text) exceed limit
                # Case2: cur_encode_text_len < limit
                new_encode_text = new_encode_text + cur_text + chunk_merge_separator
                new_origin_text = new_origin_text + cur_text + chunk_merge_separator
                if isinstance(chunk_root.get("positions"), dict) and chunk_root.get(
                    "positions"
                ).get("page_no"):
                    page_no_list = chunk_root["positions"]["page_no"]
                    if isinstance(page_no_list, list):
                        new_chunk_page_num.extend(page_no_list)
                    elif isinstance(page_no_list, int):
                        new_chunk_page_num.append(page_no_list)
                if chunk_id + 1 < len(lark_chunks):
                    # Case: not last chunk, continue
                    continue

            # cur_encode_text_len exceed limit or last chunk process
            page_num_str = ",".join(sorted([str(x) for x in set(new_chunk_page_num)]))
            new_chunk = KnowledgeChunk(
                primary_key=(
                    primary_key,
                    "#".join([doc_id, chunk_type, str(chunk_id)]),
                ),
                vector=(
                    vector_key,
                    new_encode_text[: self.config.vector_text_len_limit],
                ),
                scalars={
                    "doc_id": doc_id,
                    "chunk_id": new_chunk_id,
                    "chunk_type": chunk_type,
                    "text": new_origin_text,
                    "tos_url": tos_url,
                    "file_name": file_name,
                    "chunk_len": len(new_origin_text),
                    "full_text_len": full_text_len,
                    "title": title,
                    "page_nums": page_num_str,
                },
                knowledge_schema=self.config,
            )
            new_chunk_list.append(new_chunk)
            new_encode_text = encode_text_prefix
            new_origin_text = ""
            new_chunk_page_num = []
            new_chunk_id += 1

        return new_chunk_list

    @task()
    def load(self, doc: KnowledgeDoc) -> List[KnowledgeChunk]:
        LOGGER.info("FeishuReader load started")
        if not doc.remote_path:
            raise ValueError("doc.remote_path is empty")
        file_name = str(doc.remote_path).split("/")[-1]
        lark_chunks = self._parse_doc(doc)
        title = ""
        chunk_text = []
        for chunk in lark_chunks:
            chunk_root = json.loads(chunk)
            chunk_text.append(chunk_root["text"])
            if title == "" and chunk_root["type"] == "title":
                title = chunk_root.get("text", "")
        full_text = "\n".join(chunk_text)

        primary_key = self.config.primary_key.get("name", "")
        vector_key = self.config.vector.get("name", "")
        chunk_type = "full_text"
        chunk_id = -1
        full_text_chunk = KnowledgeChunk(
            primary_key=(
                primary_key,
                "#".join([doc.doc_id, chunk_type, str(chunk_id)]),
            ),
            vector=(
                vector_key,
                file_name + "\n" + full_text[: self.split_size - len(file_name) - 1],
            ),
            scalars={
                "doc_id": doc.doc_id,
                "chunk_id": -1,
                "chunk_type": chunk_type,
                "text": full_text,
                "tos_url": doc.remote_path,
                "file_name": file_name,
                "chunk_len": len(full_text),
                "full_text_len": len(full_text),
                "title": title,
                "page_nums": "all",
            },
            knowledge_schema=self.config,
        )
        chunks = self._rechunk(lark_chunks, full_text_chunk)
        LOGGER.info("load finished, splitted to %d chunks", len(chunks))
        LOGGER.info("FeishuReader load finished")
        return [full_text_chunk] + chunks



from typing import Any, Dict


def dict_merge(a: Dict[str, Any], b: Dict[str, Any]) -> Dict[str, Any]:
    """merge two dict recursively and return a new merged dict
    1. if two nodes in the same position are both dicts,
       merge is recursively,
    2. if two nodes in the same position are not both dicts,
       value from the second one overwrites one in first.

    Args:
        a (dict): first dict
        b (dict): second dict

    Returns:
        dict: a new merged dict
    """
    merged = dict()
    for k in set(a.keys()).union(b.keys()):
        if (k in a) and (k in b):
            if isinstance(a[k], dict) and isinstance(b[k], dict):
                merged[k] = dict_merge(a[k], b[k])
            else:
                merged[k] = b[k]
        elif k in a:
            merged[k] = a[k]
        else:  # k in b
            merged[k] = b[k]
    return merged



import asyncio
import logging
from typing import List

import numpy as np
from volcenginesdkarkruntime._exceptions import ArkAPIError

from ark.component.v3.llm.base import BaseEmbeddingLanguageModel
from ark.core.rag import KnowledgeChunk
from ark.core.task import task

logging.basicConfig(
    level=logging.INFO, format="[%(asctime)s][%(levelname)s] %(message)s"
)
LOGGER = logging.getLogger(__name__)


def norm_l2(vec: List[float]) -> List[float]:
    norm = float(np.linalg.norm(vec))
    return [v / norm for v in vec]


@task()
async def inplace_batch_update_chunk_embeddings(
    chunks: List[KnowledgeChunk],
    endpoint_id: str,
    bs: int = 20,
    retry: int = 3,
    norm: str = "none",
    vector_dim: int = 0,
    instruct: str = "",
) -> List[KnowledgeChunk]:
    """
    批量更新知识库 chunk 的 embedding
    """

    async def batch_update(chunk_batch: List[KnowledgeChunk]) -> List[KnowledgeChunk]:
        text_batch = [instruct + str(chunk.vector[1]) for chunk in chunk_batch]
        for r in range(retry):
            try:
                embeddings = await BaseEmbeddingLanguageModel(
                    endpoint_id=endpoint_id, input=text_batch
                ).abatch()
                assert len(embeddings) == len(chunk_batch), "get embeeding failed"
                for chunk, embedding in zip(chunk_batch, embeddings):
                    sliced_emb = embedding
                    if vector_dim > 0 and len(embedding) != vector_dim:
                        # need vector slice & norm
                        sliced_emb = embedding[:vector_dim]
                    if norm == "l2":
                        chunk.vector = (chunk.vector[0], norm_l2(sliced_emb))
                    else:
                        chunk.vector = (chunk.vector[0], sliced_emb)
                return chunk_batch
            except ArkAPIError as e:
                LOGGER.error("embeddings failed, retry: %s, err: %s", r, e)
        raise Exception("retry embeddings failed after retry")
        # FIXME: raise proper Exception

    # inplace update chunk.vector[1]
    # only process chunks with string type chunk.vector[1]
    unprocessed_chunks = [chunk for chunk in chunks if isinstance(chunk.vector[1], str)]
    LOGGER.info(
        "total %d chunks, %d chunks need embedding",
        len(chunks),
        len(unprocessed_chunks),
    )
    tasks = [
        batch_update(unprocessed_chunks[i : i + bs])
        for i in range(0, len(unprocessed_chunks), bs)
    ]
    await asyncio.gather(*tasks)
    return chunks



from abc import ABC, abstractmethod
from typing import Any, List, Union

from ark.core.rag import KnowledgeChunk


class BaseRetriever(ABC):
    """
    This class is used to retrieve chunks from KnowledgeBase.
    """

    @abstractmethod
    async def retrieve(
        self,
        query: Union[str, List[float]],
        retrieve_count: int,
        *args: Any,
        **kwargs: Any,
    ) -> List[KnowledgeChunk]:
        """
        Retrieve KnowledgeChunk from KnowledgeBase.
        Returns: top n similar chunks
        """



from typing import Any, List

from langchain_core.output_parsers import BaseTransformOutputParser
from pydantic.v1 import Field
from typing_extensions import Literal

from ark.component.llm.base import BaseChatLanguageModel, BaseChatPromptTemplate
from ark.component.output_parser.rag_output import RagRewriteMessageChunkOutputParser
from ark.component.prompts.rag_prompt import (
    CONDENSE_QUESTION_PROMPT_TEMPLATE,
    HYPO_ANSWER_PROMPT_TEMPLATE,
    QueryRewritePromptTemplate,
)
from ark.core._api.deprecation import deprecated
from ark.core.task import task


def _get_query_rewrite_prompt_template() -> BaseChatPromptTemplate:
    return QueryRewritePromptTemplate(template=CONDENSE_QUESTION_PROMPT_TEMPLATE)


def _get_hypo_answer_rewrite_prompt_template() -> BaseChatPromptTemplate:
    return QueryRewritePromptTemplate(template=HYPO_ANSWER_PROMPT_TEMPLATE)


def _get_default_output_parser() -> BaseTransformOutputParser:
    return RagRewriteMessageChunkOutputParser()


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.llm.rag_rewrite.RewriteGenerator",
)
class RewriteGenerator(BaseChatLanguageModel):
    name: Literal["Rewrite", "QueryRewrite", "HypoAnswer"] = "Rewrite"
    template: BaseChatPromptTemplate
    output_parser: BaseTransformOutputParser[List[str]] = Field(
        default_factory=_get_default_output_parser
    )

    def _get_prompt_template(self) -> BaseChatPromptTemplate:
        return self.template

    def _get_output_parser(self) -> BaseTransformOutputParser:
        return self.output_parser

    @task()
    async def arun(self, *args: Any, **kwargs: Any) -> bool:
        resp = await super().arun(*args, **kwargs)
        return await self.aparse_output(resp.choices[0].message.content)

    @task()
    def run(self, *args: Any, **kwargs: Any) -> bool:
        resp = super().run(*args, **kwargs)
        return self.parse_output(resp.choices[0].message.content)


class QueryRewriteGenerator(RewriteGenerator):
    name: Literal["QueryRewrite"] = "QueryRewrite"
    template: BaseChatPromptTemplate = Field(
        default_factory=_get_query_rewrite_prompt_template
    )


class HypoAnswerGenerator(RewriteGenerator):
    name: Literal["HypoAnswer"] = "HypoAnswer"
    template: BaseChatPromptTemplate = Field(
        default_factory=_get_hypo_answer_rewrite_prompt_template
    )



import logging
import os
from typing import Any, Dict, List, Optional, Tuple, Union, get_args

import pydantic
from deprecated import deprecated
from pydantic import BaseModel, field_validator, model_validator
from volcengine.viking_knowledgebase import (
    Collection,
    EmbddingModelType,
    Field,
    FieldType,
    IndexType,
    Point,
    VikingKnowledgeBaseService,
)

from ark.core.client import get_client_pool
from ark.core.rag import KnowledgeBase, KnowledgeChunk, KnowledgeDoc, KnowledgeSchema
from ark.core.utils.errorsv3 import (
    InvalidParameter,
    KnowledgeBaseError,
    MissingParameter,
)

logging.basicConfig(
    level=logging.INFO, format="[%(asctime)s][%(levelname)s] %(message)s"
)
LOGGER = logging.getLogger(__name__)

INSTANCE_TYPE = {
    "int64": int,
    "float32": float,
    "string": str,
    "bool": bool,
}

LIST_TYPE = {
    "list<string>": List[str],
    "list<int64>": List[int],
}

FIELD_ENUM_MAP = {
    "int64": FieldType.Int64,
    "string": FieldType.String,
    "bool": FieldType.Bool,
    "list<string>": FieldType.List_String,
    "list<int64>": FieldType.List_Int64,
}

ALLOWED_KEYS = {
    "chunking_strategy",
    "chunking_identifier",
    "chunk_length",
    "merge_small_chunks",
}


def _init_service(
    ak: str = "", sk: str = "", **kwargs: Any
) -> VikingKnowledgeBaseService:
    return VikingKnowledgeBaseService(
        ak=ak or os.getenv("VOLC_ACCESSKEY"), sk=sk or os.getenv("VOLC_SECRETKEY")
    )


def _default_fields_to_dict(field: Field) -> Dict[str, Any]:
    return {
        "field_name": field.field_name,
        "field_type": field.field_type,
        "default_val": field.field_val,
    }


def _is_field_type_matched(obj: Any, expected_type: str) -> bool:
    field_type = INSTANCE_TYPE.get(expected_type)
    if field_type:
        return isinstance(obj, field_type)
    list_type = LIST_TYPE.get(expected_type)
    item_type = get_args(list_type)
    return isinstance(list_type, list) and all(
        isinstance(item, item_type) for item in list_type
    )


def _is_valid_field_type(field_type: str) -> bool:
    for field in FieldType:
        if field_type == field.value:
            return True
    return False


def _merge_dicts(dict1: Dict[str, Any], dict2: Dict[str, Any]) -> Dict[str, Any]:
    merged_dict = dict1.copy()
    for key, value in dict2.items():
        if key in merged_dict:
            if isinstance(value, dict) and isinstance(merged_dict[key], dict):
                merged_dict[key] = _merge_dicts(merged_dict[key], value)
            elif isinstance(value, list) and isinstance(merged_dict[key], list):
                merged_dict[key].extend(value)
            else:
                merged_dict[key] = value
        else:
            merged_dict[key] = value
    return merged_dict


def _merge_scalars(
    scalars_collection: List[Dict[str, Any]],
    scalars_user: Optional[List[Dict[str, Any]]],
) -> List[Dict[str, Any]]:
    if scalars_user is not None:
        scalars = scalars_user
    else:
        return scalars_collection

    key = "field_name"
    seen = set()

    for _dic in scalars_collection:
        if _dic[key] not in seen:
            scalars.append(
                {
                    "field_name": _dic["field_name"],
                    "field_type": _dic["field_type"],
                    "field_value": _dic["default_val"],
                }
            )
            seen.add(_dic[key])
    return scalars


class PrimaryKey(BaseModel):
    name: str
    type: str

    @field_validator("type")
    @classmethod
    def validate_type(cls, field_type: str) -> str:
        assert _is_valid_field_type(field_type), InvalidParameter(
            f"field type {field_type} not in supported field types"
        )
        return field_type


class Vector(BaseModel):
    name: str
    model: str
    type: str
    embedding_type: Optional[str] = pydantic.Field(default="llm")
    dim: Optional[int] = None

    @model_validator(mode="before")
    @classmethod
    def validate_vector(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        field_type, dim, model, embedding_type = (
            v.get("type"),
            v.get("dim"),
            v.get("model"),
            v.get("embedding_type", "llm"),
        )

        assert field_type, MissingParameter("type")
        assert model, MissingParameter("model")
        assert embedding_type in ("vdb", "llm"), InvalidParameter("embedding_type")
        assert _is_valid_field_type(field_type), InvalidParameter(
            f"field type {field_type} not in supported field types"
        )

        if dim:
            assert 4 <= dim <= 2048, InvalidParameter("dim should be in range [4,2048]")
            assert dim % 4 == 0, InvalidParameter("dim should be dived by 4")

        return v


class Scalar(BaseModel):
    field_name: str
    field_type: str
    default_val: Optional[Any] = None

    @model_validator(mode="before")
    @classmethod
    def validate_scalar(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        field_type, default_val = v.get("field_type"), v.get("default_val")
        assert field_type, MissingParameter("field_type")
        assert _is_valid_field_type(field_type), InvalidParameter(
            f"field type {field_type} not in supported field types"
        )

        if default_val:
            assert _is_field_type_matched(default_val, field_type), InvalidParameter(
                "default_val"
            )
        return v


class VikingKnowledgeBaseSchema(KnowledgeSchema):
    @field_validator("primary_key")
    @classmethod
    def validate_primary_key(cls, v: Dict[str, str]) -> Dict[str, str]:
        try:
            PrimaryKey.model_validate(v)
            return v
        except Exception as e:
            raise e

    @field_validator("vector")
    @classmethod
    def validate_vector(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        try:
            Vector.model_validate(v)
            return v
        except Exception as e:
            raise e

    @field_validator("scalars")
    @classmethod
    def validate_scalars(cls, v: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        try:
            for scalar in v:
                Scalar.model_validate(scalar)

            return v
        except Exception as e:
            raise e

    def schema_to_field(self, data: Dict[str, Any]) -> Field:
        field_name = data.get("name")
        return Field(
            field_name=field_name,
            field_type=FIELD_ENUM_MAP.get(data.get("type", "")),
            default_val=data.get("default_val"),
            dim=data.get("dim"),
            is_primary_key=(field_name == self.primary_key.get("name")),
            pipeline_name=data.get("model")
            if data.get("embedding_type", "llm") == "vdb"
            else None,
        )

    @property
    def fields(self) -> List[Field]:
        fields: List[Field] = []
        fields.append(self.schema_to_field(self.primary_key))
        fields.append(self.schema_to_field(self.vector))

        for scalar in self.scalars:
            fields.append(self.schema_to_field(scalar))

        return fields

    def get_primary_key_data(self, field: Dict[str, Any]) -> Tuple[str, Any]:
        key_name, key_type = (
            self.primary_key.get("name"),
            self.primary_key.get("type"),
        )

        assert key_name, MissingParameter("name")
        assert key_type, MissingParameter("type")

        primary_key = field.get(key_name)
        assert primary_key, "primary key should not be empty"
        assert _is_field_type_matched(primary_key, key_type), InvalidParameter(
            f"primary_key should be type {key_type}"
        )

        return key_name, primary_key

    def get_vector_data(
        self, field: Dict[str, Any]
    ) -> Tuple[str, Union[List[Any], str]]:
        vector_name, vector_type, embedding_type = (
            self.vector.get("name"),
            self.vector.get("type"),
            self.vector.get("embedding_type"),
        )
        assert vector_name, MissingParameter("vector name")
        assert vector_type, MissingParameter("vector type")

        if embedding_type == "vdb":
            vector = field.get(vector_name, "")
            assert _is_field_type_matched(vector, vector_type), InvalidParameter(
                f"vector should be type {vector_type}"
            )

            return vector_name, vector
        # embedding_type == "llm", viking search do not provide vector field
        return vector_name, ""

    def get_scalar_data(self, field: Dict[str, Any]) -> Dict[str, Any]:
        scalar_dic: Dict[str, Any] = {}
        for scalar_schema in self.scalars:
            scalar_name, scalar_type, default_val = (
                scalar_schema.get("name"),
                scalar_schema.get("type"),
                scalar_schema.get("default_val"),
            )
            assert scalar_name, MissingParameter("scalar name")
            assert scalar_type, MissingParameter("scalar type")

            scalar = field.get(scalar_name, default_val)
            if scalar is None:
                scalar = default_val

            assert _is_field_type_matched(scalar, scalar_type), InvalidParameter(
                f"{scalar_name} should be type {scalar_type}, "
                f"type of {scalar_name} is:{type(scalar)}"
            )

            scalar_dic[scalar_name] = scalar

        assert len(scalar_dic) > 0, "scalar should not be empty"

        return scalar_dic


def _default_index(knowledge_schema: VikingKnowledgeBaseSchema) -> Dict[str, Any]:
    return {
        "index_type": IndexType.HNSW_HYBRID,
        "index_config": {
            "fields": [
                _default_fields_to_dict(field) for field in knowledge_schema.fields
            ],
            "cpu_quota": 1,
            "embedding_model": EmbddingModelType.EmbeddingModelBgeLargeZhAndM3,
        },
    }


class VikingKnowledgeBase(KnowledgeBase):
    collection: Collection
    client: VikingKnowledgeBaseService = pydantic.Field(default=_init_service())

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def _get_viking_knowledgebase_client(
        cls, ak: str = "", sk: str = ""
    ) -> VikingKnowledgeBaseService:
        client = getattr(cls, "client", None)
        if client is None:
            return _init_service(ak, sk)
        else:
            return cls.client

    def _fields_to_dict(self, fields: List[Field]) -> Dict[str, Any]:
        return {field.field_name: field.field_val for field in fields}

    def _set_chunk_scalars(self, point: Point) -> Dict[str, Any]:
        scalars = self._fields_to_dict(point.doc_info.fields)
        scalars["doc_id"] = point.doc_info.doc_id
        scalars["doc_type"] = point.doc_info.doc_type
        scalars["chunk_id"] = point.point_id
        return scalars

    def point_to_chunk(self, point: Point) -> KnowledgeChunk:
        return KnowledgeChunk(
            knowledge_schema=self.knowledge_schema,
            primary_key=("id", point.point_id),
            vector=("content", point.content),
            scalars=self._set_chunk_scalars(point),
            retrieve_score=point.score if point.score is not None else 0.0,
        )

    @classmethod
    def generate_sid(cls) -> str:
        import random
        import string
        from datetime import datetime

        TIMESTAMP_FORMAT = "%Y%m%d%H%M%S"
        RANDOM_LENGTH = 5

        timestamp = datetime.now().strftime(TIMESTAMP_FORMAT)
        random_string = "".join(
            random.choices(string.ascii_lowercase + string.digits, k=RANDOM_LENGTH)
        )
        return f"viking_knowledgebase_{timestamp}_{random_string}"

    @classmethod
    def create(
        cls,
        schema: VikingKnowledgeBaseSchema,
        description: Optional[str] = "",
        index: Dict[str, Any] = {},
        client: Optional[VikingKnowledgeBaseService] = None,
    ) -> "VikingKnowledgeBase":
        viking_knowledgebase_client: VikingKnowledgeBaseService = (
            client or cls._get_viking_knowledgebase_client()
        )
        collection_name: str = cls.generate_sid()
        default_index = _default_index(schema)
        if len(index) != 0:
            index = _merge_dicts(default_index, index)
        else:
            index = default_index
        collection: Collection = viking_knowledgebase_client.create_collection(
            collection_name=collection_name,
            index=index,
            description=description,
        )

        return cls(
            sid=collection_name,
            knowledge_schema=schema,
            collection=collection,
            client=viking_knowledgebase_client,
        )

    @classmethod
    def delete(
        cls, sid: str, client: Optional[VikingKnowledgeBaseService] = None
    ) -> None:
        viking_knowledgebase_client: VikingKnowledgeBaseService = (
            client or cls._get_viking_knowledgebase_client()
        )
        viking_knowledgebase_client.drop_collection(collection_name=sid)

    @classmethod
    def get(
        cls,
        sid: str,
        schema: VikingKnowledgeBaseSchema,
        client: Optional[VikingKnowledgeBaseService] = None,
    ) -> "VikingKnowledgeBase":
        viking_knowledgebase_client: VikingKnowledgeBaseService = (
            client or cls._get_viking_knowledgebase_client()
        )
        collection = viking_knowledgebase_client.get_collection(collection_name=sid)

        return cls(
            sid=sid,
            knowledge_schema=schema,
            collection=collection,
            client=viking_knowledgebase_client,
        )

    @deprecated  # abstract
    def search(
        self, query: Union[str, List[Any]], count: int = 10, **kwargs: Any
    ) -> List[KnowledgeChunk]:
        if isinstance(query, str):
            filter = kwargs.get("filter", {})
            dense_weight = kwargs.get("dense_weight", 0.5)
            rerank_switch = kwargs.get("rerank_switch", False)
            query_param = {"doc_filter": filter}
            points: List[Point] = self.client.search_collection(
                collection_name=self.sid,
                query=query,
                query_param=query_param,
                limit=count,
                dense_weight=dense_weight,
                rerank_switch=rerank_switch,
            )
        else:
            raise InvalidParameter("query should be either str or list")
        chunks: List[KnowledgeChunk] = []
        for point in points:
            chunks.append(self.point_to_chunk(point))

        return chunks

    def upsert_doc(self, doc: KnowledgeDoc) -> bool:
        scalars = []
        if doc.doc_scalars is not None:
            scalars = _merge_scalars(self.knowledge_schema.scalars, doc.doc_scalars)
        else:
            for _scalar in self.knowledge_schema.scalars:
                scalars.append(
                    {
                        "field_name": _scalar["field_name"],
                        "field_type": _scalar["field_type"],
                        "field_value": _scalar["default_val"],
                    }
                )

        if doc.add_type == "tos":
            self.collection.add_doc(
                add_type=doc.add_type,
                tos_path=doc.remote_path,
            )

        elif doc.add_type == "url":
            self.collection.add_doc(
                add_type="url",
                doc_id=doc.doc_id,
                doc_name=doc.doc_name,
                doc_type=doc.type.value,
                tos_path=None,
                url=doc.remote_path,
                meta=scalars,
            )

        return True

    def batch_upsert_docs(self, tos_path: str) -> bool:
        self.collection.add_doc(
            add_type="tos",
            tos_path=tos_path,
        )
        return True

    # abstract
    def get_chunk(self, chunk_id: str) -> KnowledgeChunk:
        point: Point = self.collection.get_point(chunk_id)
        return self.point_to_chunk(point)

    def batch_get_chunks(self) -> List[KnowledgeChunk]:
        points: List[Point] = self.collection.list_points()
        return [self.point_to_chunk(point) for point in points]

    # abstract
    def delete_chunks(self, chunk_id: List[str]) -> None:
        raise KnowledgeBaseError("api not support in viking_knowledgebase")

    def batch_delete_chunks(self, chunk_ids: Union[str, List[str]]) -> bool:
        raise KnowledgeBaseError("api not support in viking_knowledgebase")

    # abstract
    def upsert_chunks(self, data: List[KnowledgeChunk]) -> bool:
        raise KnowledgeBaseError("api not support in viking_knowledgebase")

    # abstract
    async def aupsert_chunks(self, data: List[KnowledgeChunk]) -> bool:
        raise KnowledgeBaseError("api not support in viking_knowledgebase")

    # abstract
    async def adelete_chunks(self, chunk_id: List[str]) -> None:
        pass

    # abstract
    async def aget_chunk(self, chunk_id: str) -> KnowledgeChunk:
        point: Point = await self.collection.async_get_point(point_id=chunk_id)
        return self.point_to_chunk(point)

    @deprecated  # abstract
    async def asearch(
        self, query: Union[str, List[Any]], count: int = 10, **kwargs: Any
    ) -> List[KnowledgeChunk]:
        if isinstance(query, str):
            filter = kwargs.get("filter", {})
            dense_weight = kwargs.get("dense_weight", 0.5)
            rerank_switch = kwargs.get("rerank_switch", False)
            query_param = {"doc_filter": filter}
            points: List[Point] = await self.client.async_search_collection(
                collection_name=self.sid,
                query=query,
                query_param=query_param,
                limit=count,
                dense_weight=dense_weight,
                rerank_switch=rerank_switch,
            )
        else:
            raise InvalidParameter("query should be either str or list")
        chunks: List[KnowledgeChunk] = []
        for point in points:
            chunks.append(self.point_to_chunk(point))
        return chunks

    async def aupsert_doc(self, doc: KnowledgeDoc) -> bool:
        scalars = _merge_scalars(self.knowledge_schema.scalars, doc.doc_scalars)
        if doc.add_type == "tos":
            await self.collection.async_add_doc(
                add_type=doc.add_type,
                tos_path=doc.remote_path,
            )

        elif doc.add_type == "url":
            await self.collection.async_add_doc(
                add_type="url",
                doc_id=doc.doc_id,
                doc_name=doc.doc_name,
                doc_type=doc.type.value,
                tos_path=None,
                url=doc.remote_path,
                scalars=scalars,
            )

        return True

    async def abatch_get_chunks(self) -> List[KnowledgeChunk]:
        points: List[Point] = await self.collection.async_list_points()
        return [self.point_to_chunk(point) for point in points]


async def get_viking_knowledge_base_client(
    ak: str = "", sk: str = ""
) -> VikingKnowledgeBaseService:
    client_pool = get_client_pool()
    client: VikingKnowledgeBaseService = client_pool.get_client("viking-knowledgebase")  # type: ignore
    if not client:
        client = VikingKnowledgeBaseService(
            ak=ak or os.getenv("VOLC_ACCESSKEY"), sk=sk or os.getenv("VOLC_SECRETKEY")
        )
    return client



import importlib
from typing import AsyncIterable, Callable

from ark.core.idl.common_protocol import RequestType, ResponseType


def load_function(
    module_name: str, func_name: str
) -> Callable[[RequestType], AsyncIterable[ResponseType]]:
    package = importlib.import_module(module_name)
    module = getattr(package, func_name)
    return module



import os
from typing import Optional

from tos.clientv2 import TosClientV2
from tos.enum import HttpMethodType
from tos.exceptions import TosClientError

from ark.core.utils.errors import OperationDenied


class TosSDKDownloader(TosClientV2):
    def __init__(
        self,
        ak: str = "",
        sk: str = "",
        endpoint: str = "https://tos-cn-beijing.volces.com",
        region: str = "cn-beijing",
        bucket_name: Optional[str] = "ark-sdk-cn-beijing",
        sdk_prefix: Optional[str] = "python_sdk",
    ):
        ak = ak or os.getenv("VOLC_ACCESSKEY") or ""
        sk = sk or os.getenv("VOLC_SECRETKEY") or ""

        super(TosSDKDownloader, self).__init__(ak, sk, endpoint, region)

        self.bucket_name = bucket_name or "ark-sdk-cn-beijing"
        self.sdk_prefix = sdk_prefix or "python_sdk"

    def get_sdk_presign_url(self, sdk_wheel: str) -> str:
        res = self.pre_signed_url(
            http_method=HttpMethodType.Http_Method_Get,
            bucket=self.bucket_name,
            key=os.path.join(self.sdk_prefix, sdk_wheel),
        )
        return res.signed_url

    def get_latest_sdk_wheel(self) -> str:
        try:
            result = self.list_objects(
                bucket=self.bucket_name, prefix=self.sdk_prefix, max_keys=1000
            )

            latest_sdk = max(result.contents, key=lambda item: item.last_modified)
            latest_sdk_wheel = latest_sdk.key.lstrip(self.sdk_prefix + "/")
            return latest_sdk_wheel
        except TosClientError as e:
            raise OperationDenied(f"List sdks from tos failed: {e}")
        except Exception:
            raise

    def download_sdk(self, sdk_version: Optional[str] = None) -> bytes:
        try:
            sdk_version = sdk_version or self.get_latest_sdk_wheel()
            object_stream = self.get_object(bucket=self.bucket_name, key=sdk_version)
            sdk_wheel = object_stream.read()
            return sdk_wheel
        except TosClientError as e:
            raise OperationDenied(f"Download sdk from tos failed: {e}")
        except Exception:
            raise



import logging
import time
from enum import Enum
from typing import Any, Union

import six

from ark.core.utils.context import get_client_reqid, get_reqid


class LoggerName:
    logger_name = "root"

    @classmethod
    def set(cls: Any, name: str) -> None:
        cls.logger_name = name

    @classmethod
    def get(cls: Any) -> str:
        return cls.logger_name


class LogIdFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        logid = getattr(record, "tags", {}).get("_reqid")
        client_reqid = getattr(record, "tags", {}).get("_client_reqid")
        if logid:
            record._logid = logid
            record._client_reqid = client_reqid
            return True
        logid = get_reqid("-")
        client_reqid = get_client_reqid("-")
        record._logid = six.ensure_text(logid)
        record._client_reqid = six.ensure_text(client_reqid)
        return True


class RpcFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        tags = getattr(record, "tags", None)
        if not tags or not isinstance(tags, dict):
            return False
        record._rpc_tags = six.ensure_text(
            " ".join(sorted(["%s:%s" % (k, v) for k, v in tags.items()]))
        )
        return True


class TimeScale(Enum):
    SECOND = 0
    MILLI_SECOND = 1
    MICRO_SECOND = 2
    NANO_SECOND = 3


class TimeDelta(object):
    """
    TimeDelta, we only store data as millisecond for the object.
    """

    __slots__ = ["_delta"]

    _k_scale_second = 1
    _k_scale_milli = 1000
    _k_scale_micro = 1000000
    _k_scale_nano = 1000000000

    _scale_map = {
        TimeScale.SECOND: _k_scale_second,
        TimeScale.MILLI_SECOND: _k_scale_milli,
        TimeScale.MICRO_SECOND: _k_scale_micro,
        TimeScale.NANO_SECOND: _k_scale_nano,
    }

    def __init__(self, delta: float, scale: TimeScale = TimeScale.SECOND) -> None:
        """
        Args:
            delta, time delta.
            scale, input delta time scale.

        """
        self._delta = delta
        if scale != TimeScale.MILLI_SECOND:
            self._delta = delta * TimeDelta._k_scale_milli / TimeDelta._scale_map[scale]

    def as_second(self) -> float:
        return self._delta / TimeDelta._k_scale_milli

    def as_micro(self) -> float:
        return self._delta * TimeDelta._k_scale_micro / TimeDelta._k_scale_milli

    def as_nano(self) -> float:
        return self._delta * TimeDelta._k_scale_nano / TimeDelta._k_scale_milli

    def __repr__(self) -> str:
        return self._delta.__repr__()

    def __add__(self, other: Union[int, float, "TimeDelta"]) -> "TimeDelta":
        if not isinstance(other, (int, float, TimeDelta)):
            return NotImplemented
        if isinstance(other, TimeDelta):
            return TimeDelta(self._delta + other._delta, scale=TimeScale.MILLI_SECOND)
        else:
            return TimeDelta(self._delta + other, scale=TimeScale.MILLI_SECOND)

    def __sub__(self, other: Union[int, float, "TimeDelta"]) -> "TimeDelta":
        if not isinstance(other, (int, float, TimeDelta)):
            return NotImplemented
        if isinstance(other, TimeDelta):
            return TimeDelta(self._delta - other._delta, scale=TimeScale.MILLI_SECOND)
        else:
            return TimeDelta(self._delta - other, scale=TimeScale.MILLI_SECOND)

    def __rsub__(self, other: Union[int, float, "TimeDelta"]) -> "TimeDelta":
        if not isinstance(other, (int, float, TimeDelta)):
            return NotImplemented
        if isinstance(other, TimeDelta):
            return TimeDelta(other._delta - self._delta, scale=TimeScale.MILLI_SECOND)
        else:
            return TimeDelta(other - self._delta, scale=TimeScale.MILLI_SECOND)

    def __truediv__(
        self, other: Union[int, float, "TimeDelta"]
    ) -> Union["TimeDelta", float]:
        if not isinstance(other, (int, float, TimeDelta)):
            return NotImplemented
        if isinstance(other, TimeDelta):
            return TimeDelta(self._delta / other._delta, scale=TimeScale.MILLI_SECOND)
        else:
            return TimeDelta(self._delta / other, scale=TimeScale.MILLI_SECOND)

    def __mul__(self, other: Union[int, float, "TimeDelta"]) -> "TimeDelta":
        if not isinstance(other, (int, float, TimeDelta)):
            return NotImplemented
        if isinstance(other, TimeDelta):
            return TimeDelta(self._delta * other._delta, scale=TimeScale.MILLI_SECOND)
        else:
            return TimeDelta(self._delta * other, scale=TimeScale.MILLI_SECOND)


class Timer(object):
    """
    A timer do record performance.
    """

    def __init__(self) -> None:
        self._time = time.perf_counter()

    def reset(self) -> "Timer":
        """
        reset base time.
        """
        self._time = time.perf_counter()
        return self

    def elapsed(self, reset: bool = True) -> float:
        """
        By default, it will reset the base time to current time.
        """
        e = TimeDelta(time.perf_counter() - self._time).as_micro()
        if reset:
            self.reset()
        return e



import json
import os
import warnings
from typing import Any, AsyncIterator, Dict, Iterator, Optional, Tuple

import aiohttp
from volcengine.ApiInfo import ApiInfo
from volcengine.base.Request import Request
from volcengine.base.Service import Service
from volcengine.Credentials import Credentials
from volcengine.maas.sse_decoder import SSEDecoder
from volcengine.maas.utils import dict_to_object
from volcengine.maas.v2.MaasService import (
    MaasException,
    new_client_sdk_request_error,
)
from volcengine.maas.v2.utils import gen_req_id
from volcengine.ServiceInfo import ServiceInfo

from ark.core.client import APIGAuthorization, AsyncSSEDecoder, Client
from ark.core.idl.maas_protocol import (
    MaasChatResponse,
)
from ark.core.utils.errors import InvalidParameter, MissingParameter


class AssistantClient(Client, Service):
    def __init__(
        self,
        host: str,
        region: str = "cn-beijing",
        api_path: str = "/api/v2/assistant/chat",
        ak: Optional[str] = None,
        sk: Optional[str] = None,
        connection_timeout: float = 60,
        socket_timeout: float = 60,
        jwt_token: Optional[str] = None,
        auto_refresh_apikey: bool = True,
        service_id: Optional[str] = None,
    ):
        if service_id and region:
            self.service_id, region = service_id, region
        elif auto_refresh_apikey:
            self.service_id, region = self.parse_apig_host(host)

        service_info = self.get_service_info(
            host, region, connection_timeout, socket_timeout
        )
        api_info = self._get_api_info(api_path)
        super(AssistantClient, self).__init__(
            service_info=service_info, api_info=api_info
        )

        try:
            self.async_session: Optional[
                aiohttp.ClientSession
            ] = aiohttp.ClientSession()
        except Exception as e:
            self.async_session = None
            warnings.warn(str(e))

        self.set_ak(ak or os.getenv("VOLC_ACCESSKEY"))
        self.set_sk(sk or os.getenv("VOLC_SECRETKEY"))

        self._setted_apikey = jwt_token
        self.authorization = None
        if auto_refresh_apikey:
            self.authorization = APIGAuthorization(
                ak=self.service_info.credentials.ak,
                sk=self.service_info.credentials.sk,
            )

    @staticmethod
    def get_service_info(
        host: str, region: str, connection_timeout: float, socket_timeout: float
    ) -> ServiceInfo:
        service_info = ServiceInfo(
            host,
            {"Accept": "application/json"},
            Credentials("", "", "ark", region),
            connection_timeout,
            socket_timeout,
            "https",
        )
        return service_info

    @staticmethod
    def _get_api_info(api_path: str) -> Dict[str, ApiInfo]:
        api_info = {
            "chat": ApiInfo("POST", api_path, {}, {}, {}),
        }
        return api_info

    @staticmethod
    def parse_apig_host(host: str) -> Tuple[str, str]:
        if not isinstance(host, str):
            raise InvalidParameter("host")

        import re

        pattern = r"([\w.-]+)\.apigateway-([\w.-])+\.volceapi\.com.*"

        matches = re.match(pattern, host)

        if matches and len(matches.group()) >= 2:
            matched_groups = matches.groups()
            service_id, region = matched_groups[0], matched_groups[1]
            return service_id, region
        else:
            raise MissingParameter("service_id and region")

    def _validate(self, api: str, req_id: str) -> None:
        if self.authorization is None and self._setted_apikey is None:
            raise new_client_sdk_request_error("no valid credential", req_id)

        if api not in self.api_info:
            raise new_client_sdk_request_error("no such api", req_id)

    def set_apikey(self, apikey: str) -> None:
        self._setted_apikey = apikey

    def chat(self, req: Dict[str, Any]) -> MaasChatResponse:
        if self.authorization:
            self.set_apikey(self.authorization.get_apikey(service_id=self.service_id))

        req["stream"] = False
        return self._request("chat", req)

    async def achat(self, req: Dict[str, Any]) -> MaasChatResponse:
        if self.authorization:
            self.set_apikey(self.authorization.get_apikey(service_id=self.service_id))

        req["stream"] = False
        return await self._arequest("chat", req)

    def stream_chat(self, req: Dict[str, Any]) -> Iterator[MaasChatResponse]:
        req_id = gen_req_id()
        self._validate("chat", req_id)

        if self.authorization:
            self.set_apikey(self.authorization.get_apikey(service_id=self.service_id))

        assert self._setted_apikey, MissingParameter(
            "please check your ak/sk or jwttoken"
        )

        try:
            req["stream"] = True
            res = self._call(
                "chat",
                req_id,
                {},
                json.dumps(req),
                self._setted_apikey,
                stream=True,
            )
            decoder = SSEDecoder(res)

            def iter() -> Iterator[MaasChatResponse]:
                for data in decoder.next():
                    if data == b"[DONE]":
                        return

                    try:
                        res = MaasChatResponse.model_validate_json(
                            str(data, encoding="utf-8")
                        )
                        res.req_id = req_id
                    except:
                        raise
                    else:
                        if res.error is not None and res.error.code_n != 0:
                            raise MaasException(
                                res.error.code_n,
                                res.error.code,
                                res.error.message,
                                req_id,
                            )
                        yield res

            return iter()
        except MaasException:
            raise
        except Exception as e:
            raise new_client_sdk_request_error(str(e))

    async def async_stream_chat(
        self, req: Dict[str, Any]
    ) -> AsyncIterator[MaasChatResponse]:
        req_id = gen_req_id()
        self._validate("chat", req_id)

        if self.authorization:
            self.set_apikey(self.authorization.get_apikey(service_id=self.service_id))

        assert self._setted_apikey, MissingParameter(
            "please check your ak/sk or jwttoken"
        )

        req["stream"] = True
        try:
            res = await self._acall(
                "chat", req_id, {}, json.dumps(req), self._setted_apikey
            )
            decoder = AsyncSSEDecoder(res.content)

            async def iter() -> AsyncIterator[MaasChatResponse]:
                async for data in decoder.next():
                    if data == b"[DONE]":
                        return

                    try:
                        res = MaasChatResponse.model_validate_json(
                            str(data, encoding="utf-8")
                        )
                        res.req_id = req_id
                    except:
                        raise
                    else:
                        if res.error is not None and res.error.code_n != 0:
                            raise MaasException(
                                res.error.code_n,
                                res.error.code,
                                res.error.message,
                                req_id,
                            )
                        yield res

            return iter()
        except Exception as e:
            raise e

    async def _arequest(
        self, api: str, req: Dict[str, Any], params: Dict[str, Any] = {}
    ) -> MaasChatResponse:
        req_id = gen_req_id()
        self._validate(api, req_id)

        if self.authorization:
            self.set_apikey(self.authorization.get_apikey(service_id=self.service_id))

        assert self._setted_apikey, MissingParameter(
            "please check your ak/sk or jwttoken"
        )

        try:
            res = await self._acall(
                api, req_id, params, json.dumps(req), self._setted_apikey
            )
            resp = dict_to_object(await res.json())
            if resp and isinstance(resp, dict):
                resp["req_id"] = req_id
            return resp
        except MaasException:
            raise
        except Exception as e:
            raise new_client_sdk_request_error(str(e), req_id)

    async def _acall(
        self, api: str, req_id: str, params: Dict[str, Any], body: str, apikey: str
    ) -> aiohttp.ClientResponse:
        assert self.async_session, ValueError("no async session")

        r = self._build_request(api, req_id, params, body, apikey)

        url = r.build()

        res = await self.async_session.post(
            url=url,
            headers=r.headers,
            data=r.body,
            timeout=aiohttp.ClientTimeout(
                sock_read=self.service_info.socket_timeout,
                connect=self.service_info.connection_timeout,
            ),
        )

        if res.status != 200:
            raw = await res.text()
            self._handle_exeption(raw, req_id)

        return res

    def _request(
        self, api: str, req: Dict[str, Any], params: Dict[str, Any] = {}
    ) -> MaasChatResponse:
        req_id = gen_req_id()

        self._validate(api, req_id)

        if self.authorization:
            self.set_apikey(self.authorization.get_apikey(service_id=self.service_id))

        assert self._setted_apikey, MissingParameter(
            "please check your ak/sk or jwttoken"
        )

        try:
            res = self._call(api, req_id, params, json.dumps(req), self._setted_apikey)
            resp = dict_to_object(res.json())
            if resp and isinstance(resp, dict):
                resp["req_id"] = req_id
            return resp

        except MaasException as e:
            raise e
        except Exception as e:
            raise new_client_sdk_request_error(str(e), req_id)

    def _call(
        self,
        api: str,
        req_id: str,
        params: Dict[str, Any],
        body: str,
        apikey: str,
        stream: bool = False,
    ) -> Request:
        r = self._build_request(api, req_id, params, body, apikey)

        url = r.build()
        res = self.session.post(
            url,
            headers=r.headers,
            data=r.body,
            timeout=(
                self.service_info.connection_timeout,
                self.service_info.socket_timeout,
            ),
            stream=stream,
        )

        if res.status_code != 200:
            raw = res.text.encode()
            res.close()
            self._handle_exeption(raw, req_id)

        return res

    def _handle_exeption(self, raw: str, req_id: str) -> None:
        try:
            resp = MaasChatResponse.model_validate_json(raw)
            resp.req_id = req_id
        except Exception:
            raise new_client_sdk_request_error(raw, req_id)
        else:
            if resp.error:
                raise MaasException(
                    resp.error.code_n, resp.error.code, resp.error.message, req_id
                )
            else:
                raise new_client_sdk_request_error(resp, req_id)

    def _build_request(
        self, api: str, req_id: str, params: Dict[str, Any], body: str, apikey: str
    ) -> Request:
        api_info = self.api_info[api]

        r = self.prepare_request(api_info, params)
        r.headers["X-Client-Request-Id"] = req_id
        r.headers["Content-Type"] = "application/json"
        r.body = body
        r.headers["Authorization"] = "Bearer " + apikey
        return r

    async def close_async_session(self) -> None:
        assert self.async_session, ValueError("no async session")
        return await self.async_session.close()



import asyncio
import base64
import hashlib
import json
import logging
import time
import traceback
from typing import Any, Dict, List, Optional, Tuple, Union, get_args

import pydantic
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding
from pydantic import BaseModel, field_validator, model_validator
from volcengine.viking_db import (
    Collection,
    Data,
    DistanceType,
    Field,
    FieldType,
    Index,
    IndexType,
    Order,
    QuantType,
    ScalarOrder,
    Text,
    VectorIndexParams,
)

from ark.component.v3.llm.base import BaseEmbeddingLanguageModel
from ark.component.v3.vectorstores.embedding import (
    inplace_batch_update_chunk_embeddings,
    norm_l2,
)
from ark.core.client import get_client_pool
from ark.core.client.vikingdb import VikingDBClient
from ark.core.rag import KnowledgeBase, KnowledgeChunk, KnowledgeSchema
from ark.core.utils.errors import InvalidParameter, MissingParameter

RAGSchema = KnowledgeSchema
RAGBase = KnowledgeBase
MemoryChunk = KnowledgeChunk

logging.basicConfig(
    level=logging.INFO, format="[%(asctime)s][%(levelname)s] %(message)s"
)
LOGGER = logging.getLogger(__name__)

INSTANCE_TYPE = {
    "int64": int,
    "float32": float,
    "string": str,
    "bool": bool,
    "text": str,
}

LIST_TYPE = {
    "list<string>": List[str],
    "list<int64>": List[int],
    "vector": List[Any],
}

FIELD_ENUM_MAP = {
    "int64": FieldType.Int64,
    "string": FieldType.String,
    "text": FieldType.Text,
    "bool": FieldType.Bool,
    "list<string>": FieldType.List_String,
    "list<int64>": FieldType.List_Int64,
    "vector": FieldType.Vector,
    "float32": FieldType.Float32,
}

FIELD_MAP = {
    FieldType.Int64: "int64",
    FieldType.String: "string",
    FieldType.Text: "text",
    FieldType.Bool: "bool",
    FieldType.List_String: "list<string>",
    FieldType.List_Int64: "list<int64>",
    FieldType.Vector: "vector",
    FieldType.Float32: "float32",
}

ORDER_MAP = {"asc": Order.Asc, "desc": Order.Desc}


def _get_vdb_client() -> VikingDBClient:
    client_pool = get_client_pool()
    client: VikingDBClient = client_pool.get_client("vdb")  # type: ignore
    if not client:
        client = VikingDBClient()
    return client


class MemoryHandler:
    @staticmethod
    def encrypt(plaintext: str, public_key: bytes) -> str:
        plaintext_byte = plaintext.encode("utf-8")
        rsa_public_key: Any = serialization.load_pem_public_key(public_key)
        ciphertext_byte = rsa_public_key.encrypt(
            plaintext_byte,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )
        ciphertext_str = base64.b64encode(ciphertext_byte).decode("utf-8")
        return ciphertext_str

    @staticmethod
    def decrypt(ciphertext: str, private_key: bytes) -> str:
        ciphertext_byte = base64.b64decode(ciphertext.encode("utf-8"))
        rsa_private_key: Any = serialization.load_pem_private_key(
            private_key, password=None
        )
        plaintext_byte = rsa_private_key.decrypt(
            ciphertext_byte,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )
        plaintext = plaintext_byte.decode("utf-8")
        return plaintext


def _generate_id(id_str: str) -> str:
    timestamp = int(time.time() * 1000)
    text = f"{timestamp}{id_str}"
    sha256 = hashlib.sha256(text.encode())
    hex_digest = sha256.hexdigest()
    random_id = str(int(hex_digest, 16))[:11]
    return random_id


def _default_index_name(collection_name: str) -> str:
    return f"{collection_name}_Index"


def _default_index(
    client: VikingDBClient, collection_name: str, custom: Dict[str, Any] = {}
) -> Index:
    filtered_fields = _default_filtered_fields()
    scalar_index = [field.field_name for field in filtered_fields]
    scalar_index.extend([field["name"] for field in custom.get("scalars", [])])
    partition_field = _default_partition_fields()[0]

    return Index(
        collection_name=collection_name,
        index_name=_default_index_name(collection_name),
        cpu_quota=custom.get("cpu_quota", 2),
        partition_by=partition_field.field_name,
        vector_index=custom.get(
            "vector_index",
            VectorIndexParams(
                distance=DistanceType.IP,
                index_type=IndexType.HNSW,
                quant=QuantType.Int8,
            ),
        ),
        scalar_index=scalar_index,
        shard_count=custom.get("shard_count", None),
        viking_db_service=client,
        stat=None,
    )


def _default_attrs_fields() -> List[Field]:
    return [
        Field(field_name="for_mem", field_type=FieldType.Bool, default_val=True),
        Field(field_name="extra", field_type=FieldType.String, default_val=""),
        Field(
            field_name="mem_inner_success", field_type=FieldType.Bool, default_val=True
        ),
        Field(
            field_name="mem_inner_enable", field_type=FieldType.Bool, default_val=True
        ),
        Field(field_name="mem_inner_attr", field_type=FieldType.Text, default_val=""),
    ]


def _default_filtered_fields() -> List[Field]:
    return [
        Field(
            field_name="foreign_id", field_type=FieldType.List_String, default_val=[]
        ),
        Field(field_name="status", field_type=FieldType.String, default_val=""),
        Field(field_name="public", field_type=FieldType.String, default_val=""),
        Field(field_name="create_ts", field_type=FieldType.Int64, default_val=0),
        Field(field_name="update_ts", field_type=FieldType.Int64, default_val=0),
        Field(field_name="invalidate_ts", field_type=FieldType.Int64, default_val=0),
        Field(field_name="pipeline", field_type=FieldType.String, default_val=""),
        Field(field_name="time_value", field_type=FieldType.Float32, default_val=1.0),
        Field(field_name="trust_value", field_type=FieldType.Float32, default_val=1.0),
    ]


def _default_partition_fields() -> List[Field]:
    return [Field(field_name="partition", field_type=FieldType.String, default_val="")]


def _default_fields() -> List[Field]:
    default_attrs_fields = _default_attrs_fields()
    default_filtered_fields = _default_filtered_fields()
    default_partition_fields = _default_partition_fields()
    default_attrs_fields.extend(default_filtered_fields)
    default_attrs_fields.extend(default_partition_fields)
    return default_attrs_fields


def _is_valid_field_type(field_type: str) -> bool:
    for field in FieldType:
        if field_type == field.value:
            return True
    return False


def _is_field_type_matched(obj: Any, expected_type: str) -> bool:
    field_type = INSTANCE_TYPE.get(expected_type)
    if field_type:
        if expected_type == "text":
            return isinstance(obj, dict) or isinstance(obj, field_type)
            # vdb lagacy
        return isinstance(obj, field_type)

    list_type = LIST_TYPE.get(expected_type)
    item_type = get_args(list_type)

    is_list_type_matched = isinstance(obj, list) and all(
        isinstance(item, item_type) for item in obj
    )
    return is_list_type_matched


async def get_embedding(query: str, endpoint_id: str) -> List[float]:
    llm = BaseEmbeddingLanguageModel(endpoint_id=endpoint_id, input=[query])
    embeddings = await llm.abatch()
    return embeddings[0] if len(embeddings) > 0 else []


class PrimaryKey(BaseModel):
    name: str
    type: str

    @field_validator("type")
    @classmethod
    def validate_type(cls, field_type: str) -> str:
        assert _is_valid_field_type(field_type), InvalidParameter(
            f"field type {field_type} not in supported field types"
        )
        return field_type


class Vector(BaseModel):
    name: str
    model: str
    type: str
    embedding_type: Optional[str] = pydantic.Field(default="llm")
    dim: Optional[int] = None

    @model_validator(mode="before")
    @classmethod
    def validate_vector(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        field_type, dim, model, embedding_type = (
            v.get("type"),
            v.get("dim"),
            v.get("model"),
            v.get("embedding_type", "llm"),
        )

        assert field_type, MissingParameter("type")
        assert model, MissingParameter("model")
        assert embedding_type in ("vdb", "llm"), InvalidParameter("embedding_type")
        assert _is_valid_field_type(field_type), InvalidParameter(
            f"field type {field_type} not in supported field types"
        )

        if dim:
            assert 4 <= dim <= 2048, InvalidParameter("dim should be in range [4,2048]")
            assert dim % 4 == 0, InvalidParameter("dim should be dived by 4")

        return v


class Scalar(BaseModel):
    name: str
    type: str
    default_val: Optional[Any] = None

    @model_validator(mode="before")
    @classmethod
    def validate_scalar(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        field_type, default_val = v.get("type"), v.get("default_val")

        assert field_type, MissingParameter("type")
        assert _is_valid_field_type(field_type), InvalidParameter(
            f"field type {field_type} not in supported field types"
        )

        if default_val:
            assert _is_field_type_matched(default_val, field_type), InvalidParameter(
                "default_val"
            )
        return v


class VikingMemorySchema(RAGSchema):
    @field_validator("primary_key")
    @classmethod
    def validate_primary_key(cls, v: Dict[str, str]) -> Dict[str, str]:
        try:
            PrimaryKey.model_validate(v)
            return v
        except Exception as e:
            raise e

    @field_validator("vector")
    @classmethod
    def validate_vector(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        try:
            Vector.model_validate(v)
            return v
        except Exception as e:
            raise e

    @field_validator("scalars")
    @classmethod
    def validate_scalars(cls, v: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        try:
            for scalar in v:
                Scalar.model_validate(scalar)

            return v
        except Exception as e:
            raise e

    def schema_to_field(self, data: Dict[str, Any]) -> Field:
        field_name = data.get("name")
        return Field(
            field_name=field_name,
            field_type=FIELD_ENUM_MAP.get(data.get("type", "")),
            default_val=data.get("default_val"),
            dim=data.get("dim"),
            is_primary_key=(field_name == self.primary_key.get("name")),
            pipeline_name=data.get("model")
            if data.get("embedding_type", "llm") == "vdb"
            else None,
        )

    @property
    def fields(self) -> List[Field]:
        fields: List[Field] = _default_fields()

        fields.append(self.schema_to_field(self.primary_key))
        fields.append(self.schema_to_field(self.vector))

        for scalar in self.scalars:
            fields.append(self.schema_to_field(scalar))

        return fields

    def get_primary_key_data(self, field: Dict[str, Any]) -> Tuple[str, Any]:
        key_name, key_type = (
            self.primary_key.get("name"),
            self.primary_key.get("type"),
        )

        assert key_name, MissingParameter("name")
        assert key_type, MissingParameter("type")

        primary_key = field.get(key_name)
        assert primary_key, "primary key should not be empty"
        assert _is_field_type_matched(primary_key, key_type), InvalidParameter(
            f"primary_key should be type {key_type}"
        )

        return key_name, primary_key

    def get_vector_data(
        self, field: Dict[str, Any]
    ) -> Tuple[str, Union[List[Any], str]]:
        vector_name, vector_type, embedding_type = (
            self.vector.get("name"),
            self.vector.get("type"),
            self.vector.get("embedding_type"),
        )
        assert vector_name, MissingParameter("vector name")
        assert vector_type, MissingParameter("vector type")

        if embedding_type == "vdb":
            vector = field.get(vector_name, "")
            assert _is_field_type_matched(vector, vector_type), InvalidParameter(
                f"vector should be type {vector_type}"
            )

            return vector_name, vector
        # embedding_type == "llm", viking search do not provide vector field
        return vector_name, ""

    def get_scalar_data(self, field: Dict[str, Any]) -> Dict[str, Any]:
        scalars = [scalar for scalar in self.scalars]
        scalars.extend(
            [
                {
                    "name": field.field_name,
                    "type": FIELD_MAP[field.field_type],
                    "default_val": field.default_val,
                }
                for field in _default_fields()
            ]
        )
        scalar_dic: Dict[str, Any] = {}
        for scalar_schema in scalars:
            scalar_name, scalar_type, default_val = (
                scalar_schema.get("name"),
                scalar_schema.get("type"),
                scalar_schema.get("default_val"),
            )
            assert scalar_name, MissingParameter("scalar name")
            assert scalar_type, MissingParameter("scalar type")

            scalar = field.get(scalar_name, default_val)
            if scalar is None:
                scalar = default_val

            # assert _is_field_type_matched(scalar, scalar_type), InvalidParameter(
            #     f"{scalar_name} should be type {scalar_type}, "
            #     f"type of {scalar_name} is:{type(scalar)}"
            # )

            scalar_dic[scalar_name] = scalar

        # assert len(scalar_dic) > 0, "scalar should not be empty"
        return scalar_dic


class VikingMemory(RAGBase):
    collection: Collection
    index: Index
    client: VikingDBClient = pydantic.Field(default_factory=_get_vdb_client)

    @property
    def memory_schema(self) -> KnowledgeSchema:
        return self.knowledge_schema

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def fields_to_chunk(self, fields: Dict[str, Any]) -> MemoryChunk:
        return MemoryChunk(
            knowledge_schema=self.knowledge_schema,
            primary_key=self.knowledge_schema.get_primary_key_data(fields),
            vector=self.knowledge_schema.get_vector_data(fields),
            scalars=self.knowledge_schema.get_scalar_data(fields),
        )

    @classmethod
    def chunk_to_data(cls, chunk: MemoryChunk, ttl: Optional[int] = None) -> Data:
        fields: Dict[str, Any] = {
            chunk.primary_key[0]: chunk.primary_key[1],
            chunk.vector[0]: chunk.vector[1],
        }

        for scalar in chunk.scalars.items():
            fields[scalar[0]] = scalar[1]

        fields["mem_inner_success"] = fields.get("mem_inner_success", True)
        fields["mem_inner_enable"] = fields.get("mem_inner_enable", True)
        fields["mem_inner_attr"] = fields.get("mem_inner_attr", "")
        return Data(fields=fields, TTL=ttl)

    @classmethod
    def generate_sid(cls) -> str:
        import random
        import string
        from datetime import datetime

        TIMESTAMP_FORMAT = "%Y%m%d%H%M%S"
        RANDOM_LENGTH = 5

        timestamp = datetime.now().strftime(TIMESTAMP_FORMAT)
        random_string = "".join(
            random.choices(string.ascii_lowercase + string.digits, k=RANDOM_LENGTH)
        )
        return f"vm_{timestamp}_{random_string}"

    @classmethod
    def create(
        cls,
        schema: VikingMemorySchema,
        description: Optional[str] = "",
        index: Dict[str, Any] = {},
        client: Optional[VikingDBClient] = None,
    ) -> "VikingMemory":
        vdb_client: VikingDBClient = client or _get_vdb_client()

        collection_name: str = cls.generate_sid()
        fields = schema.fields

        collection: Collection = vdb_client.create_collection(
            collection_name, fields, description=description
        )

        new_index = _default_index(
            client=vdb_client, collection_name=collection_name, custom=index
        )

        vdb_index: Index = vdb_client.create_index(
            collection_name=collection_name,
            index_name=new_index.index_name,
            vector_index=new_index.vector_index,
            cpu_quota=new_index.cpu_quota,
            description=new_index.description,
            partition_by=new_index.partition_by,
            scalar_index=new_index.scalar_index,
            shard_count=new_index.shard_count,
        )

        return cls(
            sid=collection_name,
            knowledge_schema=schema,
            collection=collection,
            client=vdb_client,
            index=vdb_index,
        )

    @classmethod
    def delete(cls, sid: str, client: Optional[VikingDBClient] = None) -> None:
        """
        Delete the knowledge collection
        """
        vdb_client: VikingDBClient = client or _get_vdb_client()
        return vdb_client.drop_collection(sid)

    @classmethod
    def get(
        cls,
        sid: str,
        schema: VikingMemorySchema,
        client: Optional[VikingDBClient] = None,
    ) -> "VikingMemory":
        vdb_client: VikingDBClient = client or _get_vdb_client()
        collection = vdb_client.get_collection(sid)
        if collection is not None and len(collection.indexes) > 0:
            index = collection.indexes[0]
        else:
            index = vdb_client.get_index(sid, _default_index_name(sid))
        return cls(
            sid=sid,
            knowledge_schema=schema,
            collection=collection,
            client=vdb_client,
            index=index,
        )

    def upsert_chunks(self, data: List[MemoryChunk], ttl: Optional[int] = None) -> bool:
        """
        Add data to the memory
        """
        data_list: List[Data] = []
        for d in data:
            data_list.append(self.chunk_to_data(d, ttl))

        self.collection.upsert_data(data=data_list)
        return True

    def flush(
        self,
        chunks: List[MemoryChunk],
        bs: int = 50,
        retry: int = 3,
        ttl: Optional[int] = None,
    ) -> None:
        """
        Batch flush chunk into memory
        """
        # TODO: 需要事务性支持，需要VDB实现批量刷新接口
        for i in range(0, len(chunks), bs):
            succ = False
            for r in range(retry):
                try:
                    succ = self.upsert_chunks(chunks[i : i + bs], ttl)
                except Exception as e:
                    LOGGER.error("upsert_data error, retry: %d, err: %s", r, e)
                    time.sleep(1)
                    continue
                break
            if not succ:
                raise Exception("upsert_data error after retry")
                # FIXME: raise proper Exception

    def get_chunk(self, primary_key: str) -> MemoryChunk:
        """
        Get a chunk from the memory
        """

        data: Data = self.collection.fetch_data(primary_key)

        return self.fields_to_chunk(data.fields)

    def delete_chunks(self, primary_key: Union[str, List[str]]) -> None:
        """
        Delete chunk from the memory
        """

        self.collection.delete_data(primary_key)

    def batch_delete_chunks(
        self, chunk_id: List[str], bs: int = 50, retry: int = 3
    ) -> None:
        """
        Batch delete chunk from memory
        """
        # TODO: 需要事务性支持，需要VDB实现批量删除接口
        for i in range(int(len(chunk_id) / bs) + 1):
            for r in range(retry):
                try:
                    if (i + 1) * bs <= len(chunk_id):
                        self.delete_chunks(chunk_id[i * bs : (i + 1) * bs])
                    elif i * bs < len(chunk_id):
                        self.delete_chunks(chunk_id[i * bs :])
                    else:
                        break
                except Exception as e:
                    LOGGER.error("upsert_data error, retry: %d, err: %s", r, e)
                    time.sleep(0.1)
                    continue
                break

    def search(
        self, query: Union[str, List[Any]], count: int, **kwargs: Any
    ) -> List[MemoryChunk]:
        """
        Search the memory
        """

        if isinstance(query, list):
            datas: List[Data] = self.index.search_by_vector(
                vector=query, limit=count, **kwargs
            )
        elif isinstance(query, str):
            datas = self.index.search_by_text(
                text=Text(text=query), limit=count, **kwargs
            )
        else:
            raise InvalidParameter("query should be either str or list")
        chunks: List[MemoryChunk] = []
        for data in datas:
            chunk = self.fields_to_chunk(data.fields)
            chunk.retrieve_score = data.score
            chunks.append(chunk)

        return chunks

    async def aupsert_chunks(self, data: List[MemoryChunk]) -> bool:
        """
        Add data to the memory
        TODO: refactor with async vdb client
        """

        return await asyncio.get_running_loop().run_in_executor(
            None, self.upsert_chunks, data
        )

    async def adelete_chunks(self, primary_key: List[str]) -> None:
        await self.collection.async_delete_data(primary_key)

    async def aget_chunk(self, primary_key: str) -> MemoryChunk:
        data: Data = await self.collection.async_fetch_data(primary_key)
        memory_chunk: MemoryChunk = self.fields_to_chunk(data.fields)
        extra = memory_chunk.scalars.get("extra", None)
        if extra is not None:
            memory_chunk.scalars["extra"] = json.loads(memory_chunk.scalars["extra"])
        return memory_chunk

    # 对外暴露的API
    async def aadd(
        self,
        memory: str,
        user_id: Union[str, List[str]],
        **kwargs: Any,
    ) -> str:
        sid = self.sid
        # init primary key
        primary_key = self.memory_schema.primary_key.get("name", "")
        partition = kwargs.get("partition", "L")
        user_id_str = ""
        if isinstance(user_id, list):
            user_id_str = "-".join(user_id)
        else:
            user_id_str = user_id
            user_id = [user_id]
        random_id_str = _generate_id(user_id_str)
        primary_key_str = f"{partition}-{user_id_str}-{random_id_str}"
        # init vector
        vector_key = self.memory_schema.vector.get("name", "")
        vector = memory  # before_emb
        endpoint_id = self.memory_schema.vector.get("model", "")
        vector_dim = self.memory_schema.vector.get("dim", 0)
        # init scalar
        public_key = kwargs.get("public_key", None)
        if public_key is not None:
            memory = MemoryHandler.encrypt(memory, public_key)
        extra = kwargs.get("extra", {})
        extra["content"] = memory
        current_ts = int(time.time() * 1000)
        scalars = {
            "create_ts": current_ts,
            "update_ts": current_ts,
            "extra": json.dumps(extra, ensure_ascii=False),
            "partition": partition,
            "user_id": user_id,
        }
        ttl = kwargs.get("ttl", None)
        if ttl is not None:
            scalars["invalidate_ts"] = current_ts + ttl * 1000
        kwargs.update(scalars)

        mc = MemoryChunk(
            primary_key=(primary_key, primary_key_str),
            vector=(vector_key, vector),
            scalars=kwargs,
        )

        try:
            chunks = await inplace_batch_update_chunk_embeddings(
                chunks=[mc], endpoint_id=endpoint_id, norm="l2", vector_dim=vector_dim
            )
        except Exception:
            error_string = traceback.format_exc()
            LOGGER.error(
                "请求 embedding 失败, 知识库 id: %s, 错误信息: %s", sid, error_string
            )
            return ""

        try:
            self.flush(chunks=chunks, ttl=ttl)
        except Exception:
            error_string = traceback.format_exc()
            LOGGER.error("导入失败, 记忆库 id: %s, 错误信息: %s", sid, error_string)
            return ""

        LOGGER.info("导入成功, 记忆库 id: %s", sid)
        return primary_key_str

    async def asearch(
        self, query: Union[str, List[Any]], count: int = 10, **kwargs: Any
    ) -> List[MemoryChunk]:
        endpoint_id = self.memory_schema.vector.get("model", "")
        vector_dim = self.memory_schema.vector.get("dim", 0)

        if isinstance(query, str):
            query_vector = await get_embedding(query=query, endpoint_id=endpoint_id)
        else:
            query_vector = query

        if vector_dim > 0 and len(query_vector) != vector_dim:
            query_vector = query_vector[:vector_dim]

        query_vector = norm_l2(query_vector)

        dsl_filter = kwargs.get("dsl_filter", None)
        partition = kwargs.get("partition", "L")

        datas: List[Data] = await self.index.async_search_by_vector(
            vector=query_vector,
            limit=count,
            filter=dsl_filter,
            partition=partition,
        )

        memory_chunks: List[KnowledgeChunk] = []
        for data in datas:
            chunk = self.fields_to_chunk(data.fields)
            chunk.retrieve_score = data.score
            memory_chunks.append(chunk)

        private_key = kwargs.get("private_key", None)
        for i in range(len(memory_chunks)):
            extra = memory_chunks[i].scalars.get("extra", None)
            if extra is not None:
                if private_key is not None:
                    memory_chunks[i].scalars["extra"][
                        "content"
                    ] = MemoryHandler.decrypt(
                        ciphertext=extra["content"], private_key=private_key
                    )

                memory_chunks[i].scalars["extra"] = json.loads(
                    memory_chunks[i].scalars["extra"]
                )

        return memory_chunks

    async def asearch_by_scalar(
        self, field_name: str, count: int = 10, order: str = "asc", **kwargs: Any
    ) -> List[MemoryChunk]:
        scalar_order = ScalarOrder(
            field_name=field_name, order=ORDER_MAP.get(order, "")
        )
        dsl_filter = kwargs.get("dsl_filter", None)
        partition = kwargs.get("partition", None)
        output_fields = kwargs.get("output_fields", None)
        datas: List[Data] = await self.index.async_search(
            order=scalar_order,
            limit=count,
            filter=dsl_filter,
            output_fields=output_fields,
            partition=partition,
        )

        memory_chunks: List[MemoryChunk] = []
        for data in datas:
            chunk = self.fields_to_chunk(data.fields)
            chunk.retrieve_score = data.score
            memory_chunks.append(chunk)

        private_key = kwargs.get("private_key", None)
        for i in range(len(memory_chunks)):
            extra = memory_chunks[i].scalars.get("extra", None)
            if extra is not None:
                if private_key is not None:
                    memory_chunks[i].scalars["extra"][
                        "content"
                    ] = MemoryHandler.decrypt(
                        ciphertext=extra["content"], private_key=private_key
                    )

                memory_chunks[i].scalars["extra"] = json.loads(
                    memory_chunks[i].scalars["extra"]
                )

        return memory_chunks

    async def aupdate_memory(self, memory: str, primary_key: str, **kwargs: Any) -> str:
        data: Data = await self.collection.async_fetch_data(id=primary_key)
        chunk: MemoryChunk = self.fields_to_chunk(data.fields)
        chunk.scalars.update(kwargs)
        extra: dict = kwargs.get("extra", {})

        public_key = kwargs.get("public_key", None)
        if public_key is not None:
            extra["content"] = MemoryHandler.encrypt(
                plaintext=extra["content"], public_key=public_key
            )
        extra["content"] = memory

        current_ts = int(time.time() * 1000)
        chunk.scalars.update(
            {
                "update_ts": current_ts,
                "extra": json.dumps(extra, ensure_ascii=False),
            }
        )
        ttl = kwargs.get("ttl", None)
        if ttl is not None:
            chunk.scalars["invalidate_ts"] = current_ts + ttl * 1000
        vector_key = self.memory_schema.vector.get("name", "")
        vector = memory  # before_emb
        chunk.vector = (vector_key, vector)
        try:
            endpoint_id = self.memory_schema.vector.get("model", "")
            vector_dim = self.memory_schema.vector.get("dim", 0)
            chunks = await inplace_batch_update_chunk_embeddings(
                chunks=[chunk],
                endpoint_id=endpoint_id,
                norm="l2",
                vector_dim=vector_dim,
            )
        except Exception:
            error_string = traceback.format_exc()
            LOGGER.error(
                "请求 embedding 失败, 知识库 id: %s, 错误信息: %s",
                self.sid,
                error_string,
            )
            return ""

        try:
            self.flush(chunks=chunks, ttl=ttl)
        except Exception:
            error_string = traceback.format_exc()
            LOGGER.error(
                "导入失败, 记忆库 id: %s, 错误信息: %s", self.sid, error_string
            )
            return ""

        return primary_key

    adelete_memory = adelete_chunks
    aget_memory = aget_chunk



from typing import Any, AsyncIterator, Dict, Iterator, List, Optional, Union

from pydantic import BaseModel, Field, field_validator
from typing_extensions import Literal

from ark.core._api.deprecation import deprecated
from ark.core.client import ArkClient, Client, get_client_pool
from ark.core.idl.common_protocol import ChatRole, Parameters, Tool
from ark.core.idl.maas_protocol import (
    MaasChatMessage,
    MaasChatRequest,
    MaasChatResponse,
)
from ark.core.task import task


def _get_ark_client() -> Optional[Client]:
    client_pool = get_client_pool()
    client = client_pool.get_client("chat")
    if not client:
        client = ArkClient()
    return client


def convert_messages_to_keywords(messages: List[MaasChatMessage]) -> List[str]:
    keywords: List[str] = []
    for message in messages:
        if message.role == ChatRole.SYSTEM:
            continue
        if message.role == ChatRole.FUNCTION:
            continue
        if not isinstance(message.content, str):
            continue
        keywords.append(message.content)
    return keywords


@deprecated(
    since="0.1.11", removal="0.2.0", alternative_import="ark.core.action.Action"
)
class Plugin(BaseModel):
    name: Literal["SearchIntention", "SearchSummary"]
    endpoint_id: str
    messages: List[MaasChatMessage] = Field(default_factory=list)
    parameters: Optional[Union[Parameters, Dict[str, Any]]] = None
    options: Optional[Dict[str, Any]] = None
    client: ArkClient = Field(default_factory=_get_ark_client)
    _tools: Optional[List[Tool]] = None

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    async def arun(self) -> MaasChatResponse:
        request = MaasChatRequest(
            messages=self.messages,
            tools=self._tools,
            parameters=self.parameters,
        )

        return await self._arun(self.endpoint_id, request)

    async def astream(self) -> AsyncIterator[MaasChatResponse]:
        request = MaasChatRequest(
            stream=True,
            messages=self.messages,
            tools=self._tools,
            parameters=self.parameters,
        )
        responses = await self._arun(self.endpoint_id, request)

        async for resp in responses:
            yield resp

    def run(self) -> MaasChatResponse:
        request = MaasChatRequest(
            messages=self.messages,
            tools=self._tools,
            parameters=self.parameters,
        )
        return self._run(self.endpoint_id, request)

    def stream(self) -> Iterator[MaasChatResponse]:
        request = MaasChatRequest(
            stream=True,
            messages=self.messages,
            tools=self._tools,
            parameters=self.parameters,
        )

        for resp in self._run(self.endpoint_id, request):
            yield resp

    @task()
    async def _arun(
        self, endpoint_id: str, request: MaasChatRequest
    ) -> Union[AsyncIterator[MaasChatResponse], MaasChatResponse]:
        if request.stream:
            return await self.client.async_stream_chat(
                endpoint_id=endpoint_id, req=request
            )
        else:
            return await self.client.async_chat(endpoint_id=endpoint_id, req=request)

    @task()
    def _run(
        self, endpoint_id: str, request: MaasChatRequest
    ) -> Union[Iterator[MaasChatResponse], MaasChatResponse]:
        if request.stream:
            return self.client.stream_chat(endpoint_id=endpoint_id, req=request)
        else:
            return self.client.chat(endpoint_id=endpoint_id, req=request.model_dump())

    @field_validator("messages")
    @classmethod
    def validate_messages(cls, v: List[MaasChatMessage]) -> List[MaasChatMessage]:
        if not v or len(v) == 0:
            raise ValueError("Messages are required")

        pre_role = None  # user, assistant, function, system
        for msg in v:
            if not msg.role:
                raise ValueError("Role is required")
            if msg.content is None:
                msg.content = ""

            if msg.role == ChatRole.SYSTEM:
                continue

            if msg.role == ChatRole.USER:
                assert (
                    not pre_role or pre_role == ChatRole.ASSISTANT
                ), f"{pre_role} message should not followed by a user message"

            if msg.role == ChatRole.ASSISTANT:
                assert pre_role in [
                    ChatRole.USER,
                    ChatRole.FUNCTION,
                ], "Assistant message should follow a user or function message"

            if msg.role == ChatRole.FUNCTION:
                assert (
                    pre_role == ChatRole.ASSISTANT
                ), "Function message should follow an assistant message"

            pre_role = msg.role

        return v



import json
import logging
import sys
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple

import requests
from volcengine.auth.SignerV4 import SignerV4
from volcengine.base.Request import Request
from volcengine.Credentials import Credentials

from ark.component.reader.base_reader import BaseReader
from ark.core.rag import DocType, KnowledgeDoc, KnowledgeSchema
from ark.core.utils.errors import FeishuParseException
from ark.core.utils.tos import pre_sign_tos_url

logging.basicConfig(
    level=logging.INFO, format="[%(asctime)s][%(levelname)s] %(message)s"
)
LOGGER = logging.getLogger(__name__)


class ChunkingStrategy(Enum):
    PLAIN_TEXT = 0
    PARAGRAPH = 1
    DOC_TREE = 2
    SEMANTIC = 3


class Reader(BaseReader):
    """
    Reader for knowledge base
    Need encryption
    """

    def __init__(
        self,
        config: KnowledgeSchema,
        credentials: Credentials,
        split_size: Optional[int] = None,
        region: str = "cn-beijing",
    ):
        self.config = config
        self.credentials = credentials
        self.region = region
        self.split_size = config.vector_text_len_limit
        if split_size:
            self.split_size = split_size
        if self.split_size > config.vector_text_len_limit:
            LOGGER.warning(
                "split_size:%d, > vector_text_len_limit:%d",
                split_size,
                config.vector_text_len_limit,
            )

    def _parse_res(self, response_str: str) -> Tuple[list, str]:
        # 返回的一个chunk示例：
        # {"id": 0, "type": "title", "label": "", "level": -1, "parent": -1,
        # "children": [1, 1125, 1130, 1135, 1140], "text": "xxx",
        # "positions": {"page_no": [0], "bbox": [[0.37, 0.18, 0.62, 0.21]]},
        # "table_detail": {}}
        response: dict = json.loads(response_str)
        chunking_results = response["data"]["DocChunkingResults"]
        if len(chunking_results) < 1:
            return [], ""
        chunk = chunking_results[0]
        status_string = chunk["Status"]
        status_map = json.loads(status_string)
        ret_code = status_map["code"]
        if ret_code != 0:
            LOGGER.error(
                "check response failed, ret_code: %s, Status: %s",
                ret_code,
                status_string,
            )
            raise FeishuParseException(status_string)
        return chunk["DocChunks"], ""

    def _prepare_request(
        self,
        method: str,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        doseq: int = 0,
    ) -> Request:
        if params:
            for key in params:
                if (
                    isinstance(type(params[key]), int)
                    or isinstance(type(params[key]), float)
                    or isinstance(type(params[key]), bool)
                ):
                    params[key] = str(params[key])
                elif sys.version_info[0] != 3:
                    if isinstance(type(params[key]), str):
                        params[key] = params[key].encode("utf-8")
                elif isinstance(type(params[key]), list):
                    if not doseq:
                        params[key] = ",".join(params[key])
        r = Request()
        r.set_shema("http")
        r.set_method(method)
        r.set_connection_timeout(10)
        r.set_socket_timeout(10)
        if params:
            r.set_query(params)
        r.set_path(path)
        if data is not None:
            r.set_body(json.dumps(data))
        # 生成签名
        credentials = Credentials(
            self.credentials.ak,
            self.credentials.sk,
            "air",
            "cn-north-1",
            session_token=self.credentials.session_token,
        )
        SignerV4.sign(r, credentials)
        return r

    def _get_chunking_result(
        self,
        url: str,
        doc_type: str,
        chunk_size: int = 500,
        strategy: ChunkingStrategy = ChunkingStrategy.SEMANTIC,
        parse_table_for_pdf: bool = False,
        verberos: bool = False,
    ) -> Tuple[list, str]:
        request_params = {
            "doc_infos": [
                {
                    "doc_type": doc_type,
                    "url": url,
                    "strategy": strategy.value,
                    "chunk_size": chunk_size,
                },
            ],
            "parse_table_for_pdf": parse_table_for_pdf,
        }
        method = "POST"
        path = "/api/doc_chunking"
        DOMAIN = "viking-knowledge-demo.byte-test.com"
        info_req = self._prepare_request(method, path, data=request_params)
        res = requests.request(
            method=method,
            url="https://{}{}".format(DOMAIN, path),
            headers=info_req.headers,
            data=info_req.body,
        )
        if verberos:
            LOGGER.info("res.text:%s", res.text)
        if res.status_code == 200:
            chunks, msg = self._parse_res(res.text)
            if msg:
                if verberos:
                    LOGGER.info("parse_res failed: %s", res.text)
                return [], msg
            return chunks, ""
        else:
            msg = "fetch url:{} error, res_code:{}, res_msg:{}".format(
                url, res.status_code, res.text
            )
            LOGGER.error(msg)
            return [], msg

    def _parse_doc(self, doc: KnowledgeDoc) -> List[str]:
        if doc.type == DocType.PDF:
            doc_type = "pdf"
        elif doc.type == DocType.TXT or doc.type == DocType.MARKDOWN:
            doc_type = "markdown"
        elif doc.type == DocType.DOC:
            doc_type = "doc"
        elif doc.type == DocType.DOCX:
            doc_type = "docx"
        else:
            raise ValueError(f"Invalid doc type {doc.type}")
        signed_url = pre_sign_tos_url(
            str(doc.remote_path), credentials=self.credentials, region=self.region
        )
        assert len(signed_url) > 0, "pre sign tos url failed"
        lark_chunks, _ = self._get_chunking_result(
            signed_url,
            doc_type,
            chunk_size=self.split_size,
            strategy=ChunkingStrategy.SEMANTIC,
        )
        return lark_chunks



import copy
import json
import os
import warnings
from typing import Any, AsyncIterator, Dict, Iterator, Optional, Set, Type, Union

import aiohttp
from pydantic import BaseModel
from volcengine.ApiInfo import ApiInfo
from volcengine.auth.SignerV4 import SignerV4
from volcengine.base.Request import Request
from volcengine.maas.v2.MaasService import (
    MaasException,
    MaasService,
    SSEDecoder,
    gen_req_id,
    json_to_object,
    new_client_sdk_request_error,
)

from ark.core.client import Client
from ark.core.client.authorization import ArkAuthorization
from ark.core.client.sse import AsyncSSEDecoder
from ark.core.idl.maas_protocol import (
    MaasChatRequest,
    MaasChatResponse,
    MaasClassificationRequest,
    MaasClassificationResponse,
    MaasEmbeddingsRequest,
    MaasEmbeddingsResponse,
    MaasTokenizeRequest,
    MaasTokenizeResponse,
    SearchRequest,
    SearchResponse,
)
from ark.core.utils.context import get_extra_headers, get_reqid
from ark.core.utils.errors import (
    GPTException,
)


class ArkClient(Client, MaasService):
    def __init__(
        self,
        host: str = "maas-api.ml-platform-cn-beijing.volces.com",
        region: str = "cn-beijing",
        ak: Optional[str] = None,
        sk: Optional[str] = None,
        connection_timeout: float = 60,
        socket_timeout: float = 60,
        apikey: Optional[str] = None,
        auto_refresh_apikey: bool = False,
        max_retries: int = 1,
        **kwargs: Any,
    ):
        super().__init__(
            host=host,
            region=region,
            connection_timeout=connection_timeout,
            socket_timeout=socket_timeout,
            **kwargs,
        )
        ak = ak or os.getenv("VOLC_ACCESSKEY")
        sk = sk or os.getenv("VOLC_SECRETKEY")
        self.set_ak(ak)
        self.set_sk(sk)

        self.add_action_info()

        self._setted_apikey: Optional[str] = None
        if apikey:
            self.set_apikey(apikey=apikey)

        self.authorization = None
        try:
            self.async_session: Optional[
                aiohttp.ClientSession
            ] = aiohttp.ClientSession()
        except Exception as e:
            self.async_session = None
            warnings.warn(str(e))

        if auto_refresh_apikey:
            self.endpoint_ids: Set[str] = set()
            self.authorization = ArkAuthorization(ak=ak, sk=sk)

        self.max_retries = max_retries

    def add_action_info(self) -> Dict[str, ApiInfo]:
        action_info = {
            "search": ApiInfo("POST", "/api/v2/action/Search", {}, {}, {}),
            "search_intention": ApiInfo(
                "POST", "/api/v2/action/SearchIntention", {}, {}, {}
            ),
            "search_summary": ApiInfo(
                "POST", "/api/v2/action/SearchSummary", {}, {}, {}
            ),
        }

        self.api_info.update(action_info)
        return self.api_info

    def set_apikey(
        self, endpoint_id: Optional[str] = None, apikey: Optional[str] = None
    ) -> None:
        if apikey:
            self._setted_apikey = apikey
            return

        if endpoint_id and self.authorization:
            self.endpoint_ids.add(endpoint_id)
            self._setted_apikey = self.authorization.get_apikey(
                resource_ids=self.endpoint_ids
            )

    def _validate(self, api: str, req_id: str) -> None:
        if self.service_info.credentials is None or (
            self._setted_apikey is None
            and (
                self.service_info.credentials.sk is None
                or self.service_info.credentials.ak is None
            )
        ):
            raise new_client_sdk_request_error("no valid credential", req_id)

        if api not in self.api_info:
            raise new_client_sdk_request_error("no such api", req_id)

    def chat(self, endpoint_id: str, req: Dict[str, Any]) -> MaasChatResponse:
        if self.authorization:
            self.set_apikey(endpoint_id=endpoint_id)

        req["stream"] = False
        return self._request(endpoint_id, "chat", req)

    def stream_chat(
        self, endpoint_id: str, req: Union[MaasChatRequest, Dict[str, Any]]
    ) -> Iterator[MaasChatResponse]:
        if self.authorization:
            self.set_apikey(endpoint_id=endpoint_id)

        req_id: str = get_reqid() or gen_req_id()
        self._validate("chat", req_id)
        try:
            if isinstance(req, MaasChatRequest):
                req.stream = True
                data: str = req.model_dump_json(exclude_unset=True, exclude_none=True)
            else:
                req["stream"] = True
                data = json.dumps(req)

            res = self._call(
                endpoint_id,
                "chat",
                req_id,
                {},
                data,
                self._setted_apikey,
            )
            decoder = SSEDecoder(res)

            def iter() -> Iterator[MaasChatResponse]:
                for data in decoder.next():
                    if data == b"[DONE]":
                        return

                    try:
                        res = MaasChatResponse.model_validate_json(
                            str(data, encoding="utf-8")
                        )
                        res.req_id = req_id
                    except:
                        raise
                    else:
                        if res.error:
                            if res.error and res.error.code_n:
                                raise GPTException(
                                    code_n=res.error.code_n,
                                    message=res.error.message,
                                )
                        yield res

            return iter()
        except MaasException:
            raise
        except Exception as e:
            raise new_client_sdk_request_error(str(e))

    async def async_stream_chat(
        self, endpoint_id: str, req: Union[MaasChatRequest, Dict[str, Any]]
    ) -> AsyncIterator[MaasChatResponse]:
        if self.authorization:
            self.set_apikey(endpoint_id=endpoint_id)

        req_id = get_reqid() or gen_req_id()
        self._validate("chat", req_id)

        if isinstance(req, MaasChatRequest):
            req.stream = True
            data = req.model_dump_json(exclude_unset=True, exclude_none=True)
        else:
            req["stream"] = True
            data = json.dumps(req)

        try:
            res = await self._acall(
                endpoint_id, "chat", req_id, {}, data, self._setted_apikey
            )
            decoder = AsyncSSEDecoder(res.content)

            async def iter() -> AsyncIterator[MaasChatResponse]:
                async for data in decoder.next():
                    if data == b"[DONE]":
                        return

                    try:
                        res = MaasChatResponse.model_validate_json(
                            str(data, encoding="utf-8")
                        )
                        res.req_id = req_id
                    except:
                        raise
                    else:
                        if res.error and res.error.code_n:
                            raise GPTException(
                                code_n=res.error.code_n,
                                message=res.error.message,
                            )
                        yield res

            return iter()
        except Exception as e:
            raise e

    def tokenize(self, endpoint_id: str, req: Dict[str, Any]) -> MaasTokenizeResponse:
        if self.authorization:
            self.set_apikey(endpoint_id=endpoint_id)

        return self._request(endpoint_id, "tokenization", req)

    def classification(
        self, endpoint_id: str, req: Dict[str, Any]
    ) -> MaasClassificationResponse:
        if self.authorization:
            self.set_apikey(endpoint_id=endpoint_id)
        return self._request(endpoint_id, "classification", req)

    def embeddings(
        self, endpoint_id: str, req: Dict[str, Any]
    ) -> MaasEmbeddingsResponse:
        if self.authorization:
            self.set_apikey(endpoint_id=endpoint_id)
        return self._request(endpoint_id, "embeddings", req)

    async def async_chat(
        self, endpoint_id: str, req: Union[MaasChatRequest, Dict[str, Any]]
    ) -> MaasChatResponse:
        if isinstance(req, MaasChatRequest):
            req.stream = False
        else:
            req["stream"] = False
        return await self._arequest(endpoint_id, "chat", req, cls=MaasChatResponse)

    async def async_tokenize(
        self, endpoint_id: str, req: Union[MaasTokenizeRequest, Dict[str, Any]]
    ) -> MaasTokenizeResponse:
        return await self._arequest(
            endpoint_id, "tokenization", req, cls=MaasTokenizeResponse
        )

    async def async_classification(
        self, endpoint_id: str, req: Union[MaasClassificationRequest, Dict[str, Any]]
    ) -> MaasClassificationResponse:
        return await self._arequest(
            endpoint_id, "classification", req, cls=MaasClassificationResponse
        )

    async def async_embeddings(
        self, endpoint_id: str, req: Union[MaasEmbeddingsRequest, Dict[str, Any]]
    ) -> MaasEmbeddingsResponse:
        return await self._arequest(
            endpoint_id, "embeddings", req, cls=MaasEmbeddingsResponse
        )

    async def async_search(
        self, endpoint_id: str, request: SearchRequest
    ) -> SearchResponse:
        return await self._arequest(endpoint_id, "search", request, cls=SearchResponse)

    async def _arequest(
        self,
        endpoint_id: str,
        api: str,
        req: Union[BaseModel, Dict[str, Any]],
        params: Dict[str, Any] = {},
        cls: Type[BaseModel] = MaasChatResponse,
    ) -> Any:
        req_id = get_reqid() or gen_req_id()
        if self.authorization:
            self.set_apikey(endpoint_id=endpoint_id)

        self._validate(api, req_id)

        try:
            data: str = (
                req.model_dump_json(exclude_unset=True, exclude_none=True)
                if isinstance(req, BaseModel)
                else json.dumps(req)
            )

            res = await self._acall(
                endpoint_id,
                api,
                req_id,
                params,
                data,
                self._setted_apikey,
            )

            resp = cls.model_validate(await res.json())
            if "req_id" in cls.model_fields:
                resp.req_id = req_id  # type: ignore
            return resp

        except (MaasException, GPTException):
            raise
        except Exception as e:
            raise new_client_sdk_request_error(str(e), req_id)

    async def _acall(
        self,
        endpoint_id: str,
        api: str,
        req_id: str,
        params: Dict[str, Any],
        body: str,
        apikey: Optional[str] = None,
    ) -> aiohttp.ClientResponse:
        r = self._build_request(endpoint_id, api, req_id, params, body, apikey)

        url = r.build()
        last_err = None
        for i in range(self.max_retries):
            try:
                if not self.async_session:
                    self.async_session = aiohttp.ClientSession()

                res = await self.async_session.post(
                    url=url,
                    headers=r.headers,
                    data=r.body,
                    timeout=aiohttp.ClientTimeout(
                        total=-1,
                        sock_read=self.service_info.socket_timeout,
                        connect=self.service_info.connection_timeout,
                    ),
                )
                if res.status != 200:
                    raw: Union[str, bytes] = await res.text()
                    try:
                        resp = json_to_object(
                            raw if isinstance(raw, str) else raw.decode("utf-8"),
                            req_id=req_id,
                        )  # type: ignore
                    except Exception:
                        raise new_client_sdk_request_error(raw, req_id)
                    else:
                        if resp.error:
                            raise GPTException(
                                code_n=resp.error.code_n, message=resp.error.message
                            )
                        else:
                            raise new_client_sdk_request_error(resp, req_id)
                return res
            except Exception as e:
                last_err = e

        if last_err:
            raise last_err
        else:
            raise new_client_sdk_request_error(
                f"retry {self.max_retries} times failed", req_id
            )

    def _build_request(
        self,
        endpoint_id: str,
        api: str,
        req_id: str,
        params: Dict[str, Any],
        body: str,
        apikey: Optional[str] = None,
    ) -> Request:
        api_info = copy.deepcopy(self.api_info[api])
        api_info.path = api_info.path.format(endpoint_id=endpoint_id)

        r = self.prepare_request(api_info, params)
        r.headers["x-tt-logid"] = req_id
        r.headers["Content-Type"] = "application/json"
        r.body = body

        if apikey is None:
            SignerV4.sign(r, self.service_info.credentials)
        elif apikey is not None:
            r.headers["Authorization"] = "Bearer " + apikey
        extra_headers = get_extra_headers()
        if extra_headers:
            r.headers.update(extra_headers)

        return r



import asyncio
import copy
import logging
import time
from typing import Any, Callable, Dict, List, Optional

from ark.component.llm.base import (
    BaseChatLanguageModel,
    BaseTokenizeLanguageModel,
)
from ark.component.prompts.rag_prompt import (
    AbstractAugmentPromptTemplate,
    AugmentPromptTemplate,
    abstract_augment_prompt_template,
    faq_augment_prompt_template,
    hypo_queries_augment_prompt_template,
    summary_augment_prompt_template,
)
from ark.core._api.deprecation import deprecated
from ark.core.idl.maas_protocol import MaasChatMessage
from ark.core.rag import KnowledgeChunk, KnowledgeSchema
from ark.core.task import task
from ark.core.utils.prompt import format_maas_prompts

logging.basicConfig(
    level=logging.INFO, format="[%(asctime)s][%(levelname)s] %(message)s"
)
LOGGER = logging.getLogger(__name__)


@task()
async def llm_chat(
    endpoint_id: str,
    parameters: Dict[str, Any],
    messages: List[MaasChatMessage],
    retry: int = 1,
) -> str:
    for r in range(retry):
        try:
            resp = await BaseChatLanguageModel(
                endpoint_id=endpoint_id, parameters=parameters, messages=messages
            ).arun()
            if resp.error is not None:
                if resp.error.code_n in [1709820, 1709821]:
                    # RPM 超限错误码 1709820, TPM 超限错误码 1709821
                    await asyncio.sleep(5)
            assert isinstance(resp.choices[0].message.content, str)
            assert len(resp.choices[0].message.content) > 0
            return resp.choices[0].message.content
        except Exception as e:
            LOGGER.error("maas chat error, retry: %s, err: %s", r, e)
            await asyncio.sleep(1)
            continue
    raise Exception(f"maas chat error after {retry} trials")


@task()
async def llm_tokenize(
    endpoint_id: str,
    text: str,
    retry: int = 1,
) -> List[str]:
    """
    deprecated.
    """
    for r in range(retry):
        try:
            # FIXME：tokenize 接口的 "text" 限长不能超过 32768，MaaS fix 后就不用分段
            # text_tokens = await BaseTokenizeLanguageModel(
            #     endpoint_id=endpoint_id,
            #     text=text,
            # ).arun()
            tasks = [
                BaseTokenizeLanguageModel(
                    endpoint_id=endpoint_id,
                    text=text[i : i + 32768],
                ).arun()
                for i in range(0, len(text), 32768)
            ]
            resps = await asyncio.gather(*tasks)
            text_tokens: List[str] = []
            for resp in resps:
                assert resp.tokens is not None, "tokenize failed"
                text_tokens.extend(resp.tokens)
            return text_tokens
        except Exception as e:
            LOGGER.error("maas chat error, retry: %s, err: %s", r, e)
            await asyncio.sleep(1)
            continue
    raise Exception(f"maas tokenize error after {retry} trials")


@task()
async def llm_tokenize_v2(
    endpoint_id: str,
    text: str,
    retry: int = 1,
) -> List[str]:
    """
    use offset_mapping to generate text tokens
    """
    for r in range(retry):
        try:
            # tokenize 接口的 "text" 限长不能超过 131072
            split_size = 131072
            tasks = [
                BaseTokenizeLanguageModel(
                    endpoint_id=endpoint_id,
                    text=text[i : i + split_size],
                ).arun()
                for i in range(0, len(text), split_size)
            ]
            resps = await asyncio.gather(*tasks)
            text_tokens: List[str] = []
            last_left, last_right = -1, -1
            for i, resp in enumerate(resps):
                assert resp.offset_mapping is not None, "tokenize failed"
                for offset in resp.offset_mapping:
                    left = i * split_size + offset[0]
                    right = left + offset[1] - offset[0]
                    if left != last_left and right != last_right:
                        text_tokens.append(text[last_left:last_right])
                        last_left, last_right = left, right
                    else:
                        text_tokens.append("")
                        # multi tokens represent one utf-8 char
                        # use "" to fill
            text_tokens.append(text[last_left:last_right])
            text_tokens.pop(0)  # pop first "" from text[-1:-1]
            return text_tokens
        except Exception as e:
            LOGGER.error("maas chat error, retry: %s, err: %s", r, e)
            await asyncio.sleep(1)
            continue
    raise Exception(f"maas tokenize error after {retry} trials")


@task()
async def generate_abstract(
    config: KnowledgeSchema,
    full_text_chunk: KnowledgeChunk,
    endpoint_id: str,
    parameters: Dict[str, Any],
    field_name: str = "text",
    template: Optional[AbstractAugmentPromptTemplate] = None,
    overlap_tokens: int = 200,
) -> KnowledgeChunk:
    LOGGER.info("generate_abstract started")
    start_time = time.time()
    if not template:
        template = abstract_augment_prompt_template
    # Template Tokens
    tpl_tokens = await llm_tokenize_v2(
        endpoint_id=endpoint_id, text=template.template_str, retry=2
    )
    # full_text tokens
    full_text = full_text_chunk.scalars.get(field_name, None)
    assert full_text is not None, f"{field_name} not found in full_text_chunk.scalar"
    full_text_tokens = await llm_tokenize_v2(
        endpoint_id=endpoint_id, text=full_text, retry=2
    )

    # calculate split_size for full_text
    split_limit = parameters["max_prompt_tokens"] - len(tpl_tokens) - 2 * overlap_tokens
    split_num = len(full_text_tokens) // split_limit + 1
    split_size = min(
        len(full_text_tokens) // split_num, split_limit - parameters["max_new_tokens"]
    )

    LOGGER.info(
        "full_text_tokens %d, split_num: %d, augment may be slow",
        len(full_text_tokens),
        split_num,
    )

    # llm abstract
    abstract = "无"
    for i in range(0, len(full_text_tokens), split_size):
        cur_text = "".join(
            full_text_tokens[
                max(0, i - overlap_tokens) : i + split_size + overlap_tokens
            ]
        )
        messages = format_maas_prompts(
            template=template,
            chat_messages=[],
            last_abstract=abstract,
            reference=cur_text,
        )
        abstract_content = await llm_chat(
            endpoint_id=endpoint_id, parameters=parameters, messages=messages, retry=2
        )
        if len(abstract_content) == 0:
            continue
        abstract = abstract_content

    # build abstract chunk
    abstract_scalars = copy.deepcopy(full_text_chunk.scalars)
    chunk_type = "abstract"
    doc_id = full_text_chunk.scalars["doc_id"]
    chunk_id = full_text_chunk.scalars["chunk_id"]
    abstract_scalars.update(
        {
            "text": abstract,
            "chunk_type": chunk_type,
            "chunk_len": len(abstract),
        }
    )
    encode_abstract = abstract[: config.vector_text_len_limit]
    abstract_chunk = KnowledgeChunk(
        primary_key=(
            full_text_chunk.primary_key[0],
            "#".join([doc_id, chunk_type, str(chunk_id)]),
        ),
        vector=(full_text_chunk.vector[0], encode_abstract),
        scalars=abstract_scalars,
        knowledge_schema=full_text_chunk.knowledge_schema,
    )
    elapsed = time.time() - start_time
    LOGGER.info("generate_abstract finished, elapsed: %.2f seconds", elapsed)
    return abstract_chunk


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.llm.rag_rewrite.generate_augment",
)
@task()
async def generate_agument(
    config: KnowledgeSchema,
    chunks: List[KnowledgeChunk],
    endpoint_id: str,
    parameters: Dict[str, Any],
    field_name: str = "text",
    field_prefix: str = "",
    replace_field_with_augment: bool = False,
    template: Optional[AugmentPromptTemplate] = None,
    augment_type: str = "augment",
    augment_parser: Callable[[str], List[str]] = lambda s: [s],
    augment_num_limit: int = 1,
    parallel: int = 5,
) -> List[KnowledgeChunk]:
    LOGGER.info("%s started", augment_type)
    # run augment
    augment_results = []
    for i in range(0, len(chunks), parallel):
        tasks = []
        for chunk in chunks[i : i + parallel]:
            cur_text = chunk.scalars.get(field_name, None)
            assert cur_text, f"{field_name} not found in chunk.scalar"
            messages = format_maas_prompts(
                template=template,
                chat_messages=[],
                reference=f"{field_prefix}{cur_text}",
            )
            tasks.append(
                llm_chat(
                    endpoint_id=endpoint_id,
                    parameters=parameters,
                    messages=messages,
                    retry=2,
                )
            )
        augment_results.extend(await asyncio.gather(*tasks))
        LOGGER.info(
            "%d chunks augmented, total %d chunks", len(augment_results), len(chunks)
        )
    # build augment chunks
    augment_chunks = []
    for chunk, augment_result in zip(chunks, augment_results):
        augments = augment_parser(augment_result)
        for augment in augments[:augment_num_limit]:
            augment_scalars = copy.deepcopy(chunk.scalars)
            doc_id = chunk.scalars["doc_id"]
            chunk_id = chunk.scalars["chunk_id"]
            augment_scalars["chunk_type"] = augment_type
            if replace_field_with_augment:
                augment_scalars.update({field_name: augment, "chunk_len": len(augment)})
            encode_augment = augment[: config.vector_text_len_limit]
            augment_chunks.append(
                KnowledgeChunk(
                    primary_key=(
                        chunk.primary_key[0],
                        "#".join([doc_id, augment_type, str(chunk_id)]),
                    ),
                    vector=(chunk.vector[0], encode_augment),
                    scalars=augment_scalars,
                    knowledge_schema=chunk.knowledge_schema,
                )
            )

    LOGGER.info("%s finished", augment_type)
    return augment_chunks


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.llm.rag_rewrite.generate_summary_augment",
)
@task()
async def generate_summary_agument(
    config: KnowledgeSchema,
    chunks: List[KnowledgeChunk],
    endpoint_id: str,
    field_name: str = "text",
    field_prefix: str = "",
    parameters: Optional[Dict[str, Any]] = None,
    augment_num_limit: int = 1,
    parallel: int = 5,
) -> List[KnowledgeChunk]:
    return await generate_agument(
        config=config,
        chunks=chunks,
        endpoint_id=endpoint_id,
        field_name=field_name,
        field_prefix=field_prefix,
        replace_field_with_augment=True,
        template=summary_augment_prompt_template,
        parameters=parameters,
        augment_type="summary_augment",
        augment_parser=lambda s: [s] if len(s) > 0 and "无法提取" not in s else [],
        augment_num_limit=augment_num_limit,
        parallel=parallel,
    )


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.llm.rag_rewrite.generate_hypo_queries_augment",
)
@task()
async def generate_hypo_queries_agument(
    config: KnowledgeSchema,
    chunks: List[KnowledgeChunk],
    endpoint_id: str,
    field_name: str = "text",
    field_prefix: str = "",
    parameters: Optional[Dict[str, Any]] = None,
    augment_num_limit: int = 1,
    parallel: int = 5,
) -> List[KnowledgeChunk]:
    return await generate_agument(
        config=config,
        chunks=chunks,
        endpoint_id=endpoint_id,
        field_name=field_name,
        field_prefix=field_prefix,
        replace_field_with_augment=False,
        template=hypo_queries_augment_prompt_template,
        parameters=parameters,
        augment_type="hypo_queries_augment",
        augment_parser=lambda s: [s] if len(s) > 0 and "无法提取" not in s else [],
        augment_num_limit=augment_num_limit,
        parallel=parallel,
    )


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.llm.rag_rewrite.generate_faq_augment",
)
@task()
async def generate_faq_agument(
    config: KnowledgeSchema,
    chunks: List[KnowledgeChunk],
    endpoint_id: str,
    field_name: str = "text",
    field_prefix: str = "",
    parameters: Optional[Dict[str, Any]] = None,
    augment_num_limit: int = 3,
    parallel: int = 5,
) -> List[KnowledgeChunk]:
    def parse_faq_content(content: str) -> List[str]:
        faqs = []
        for qa_pair in content.split("\n\n"):
            lines = qa_pair.split("\n")
            if len(lines) == 2:
                faqs.append(f"问题：{lines[0]}\n回答：{lines[1]}")
        return faqs

    return await generate_agument(
        config=config,
        chunks=chunks,
        endpoint_id=endpoint_id,
        field_name=field_name,
        field_prefix=field_prefix,
        replace_field_with_augment=True,
        template=faq_augment_prompt_template,
        parameters=parameters,
        augment_type="faq_augment",
        augment_parser=parse_faq_content,
        augment_num_limit=augment_num_limit,
        parallel=parallel,
    )



import json
import os
from typing import Any, Dict, List, Optional

from pydantic import BaseModel
from volcengine.ApiInfo import ApiInfo
from volcengine.base.Service import Service
from volcengine.Credentials import Credentials
from volcengine.ServiceInfo import ServiceInfo

from ark.core.client.base import Client


class InvalidOperationError(Exception):
    def __init__(
        self, message: str, code: str = "InvalidOperation", data: dict = {}
    ) -> None:
        self.code = code
        self.message = message
        self.data = data

    def is_invalid_image_status(self) -> bool:
        return "Source image sync is in Running status" in self.message


class Env(BaseModel):
    Key: str
    Value: str


class Item(BaseModel):
    Key: str
    Value: List[str]


class Filter(BaseModel):
    Item: Item


class Function(BaseModel):
    Id: str
    Name: str
    Description: Optional[str] = None
    Runtime: str
    ExclusiveMode: bool
    RequestTimeout: int
    MaxReplicas: Optional[int] = None
    MaxConcurrency: int
    MemoryMB: int
    Envs: Optional[List[Env]] = None
    CreationTime: str
    LastUpdateTime: str


class ReleaseFunctionResponse(BaseModel):
    FunctionId: str
    Status: str
    StatusMessage: Optional[str] = None
    StableRevisionNumber: Optional[int] = None
    ErrorCode: Optional[str] = None


class VeFaaS(Client, Service):
    def __init__(
        self,
        host: str = "open.volcengineapi.com",
        region: str = "cn-beijing",
        ak: Optional[str] = None,
        sk: Optional[str] = None,
        connection_timeout: int = 300,
        socket_timeout: int = 300,
        **kwargs: Any,
    ):
        super().__init__(
            service_info=self.get_service_info(
                host, region, connection_timeout, socket_timeout
            ),
            api_info=self.get_api_info(),
            **kwargs,
        )
        self.set_ak(ak or os.getenv("VOLC_ACCESSKEY"))
        self.set_sk(sk or os.getenv("VOLC_SECRETKEY"))

    @staticmethod
    def get_service_info(
        host: str,
        region: str,
        connection_timeout: int,
        socket_timeout: int,
    ) -> ServiceInfo:
        service_info = ServiceInfo(
            host,
            {"Accept": "application/json"},
            Credentials("", "", "vefaas", region),
            connection_timeout,
            socket_timeout,
            "https",
        )
        return service_info

    @staticmethod
    def get_api_info() -> Dict[str, ApiInfo]:
        api_info = {
            "ListFunctions": ApiInfo(
                "POST",
                "/ListFunctions",
                {"Action": "ListFunctions", "Version": "2021-03-03"},
                {},
                {},
            ),
            "Release": ApiInfo(
                "POST",
                "/Release",
                {"Action": "Release", "Version": "2021-03-03"},
                {},
                {},
            ),
            "CreateFunction": ApiInfo(
                "POST",
                "/CreateFunction",
                {"Action": "CreateFunction", "Version": "2021-03-03"},
                {},
                {},
            ),
            "UpdateFunction": ApiInfo(
                "POST",
                "/UpdateFunction",
                {"Action": "UpdateFunction", "Version": "2021-03-03"},
                {},
                {},
            ),
            "CreateApigTrigger": ApiInfo(
                "POST",
                "/CreateApigTrigger",
                {"Action": "CreateApigTrigger", "Version": "2021-03-03"},
                {},
                {},
            ),
        }
        return api_info

    def get_function_by_name(self, name: str) -> Function:
        try:
            item = Item(Key="Name", Value=[name])
            filter = Filter(Item=item).model_dump()
            req = {
                "Filters": [filter],
            }
            res = self.json("ListFunctions", {}, json.dumps(req))
            res = json.loads(res)
            functions = res.get("Result", {}).get("Items", [])
            assert len(functions) > 0, "No function found"
            return Function.model_validate(functions[0])
        except Exception:
            raise

    def update_function_image(self, function_id: str, source: str) -> Function:
        try:
            req = {"Id": function_id, "SourceType": "image", "Source": source}
            res = self.json("UpdateFunction", {}, json.dumps(req))
            res = json.loads(res)
            function = res.get("Result", {})

            return Function.model_validate(function)
        except Exception:
            raise

    def release_function(  # type: ignore
        self,
        function_name: str,
        function_id: Optional[str] = None,
        revision_number: int = 0,
    ) -> ReleaseFunctionResponse:
        try:
            req = {
                "FunctionId": function_id
                or self.get_function_by_name(function_name).Id,
                "RevisionNumber": revision_number,
            }
            res = self.json("Release", {}, json.dumps(req))
            res = json.loads(res)
            resp = res.get("Result", {})

            return ReleaseFunctionResponse.model_validate(resp)
        except Exception as e:
            resp_err = json.loads(e.args[0])
            error = resp_err.get("ResponseMetadata", {}).get("Error", {})
            if error.get("Code", "") == "InvalidOperation":
                raise InvalidOperationError(error.get("Message", ""))



import asyncio
import time
from typing import Any, AsyncIterable, Dict, Optional, Tuple, Union

from pydantic import BaseModel, Field
from volcenginesdkarkruntime._exceptions import (
    ArkAPIError,
)

from ark.core.client import ArkActionClient, Client, get_client_pool
from ark.core.idl.ark_protocol import (
    ArkToolMeta,
    ArkToolRequest,
    ArkToolResponse,
)
from ark.core.idl.common_protocol import (
    ActionDetails,
    ExceptionInfo,
    ToolDetails,
    ToolOutput,
    ToolOutputType,
)
from ark.core.task.task import task
from ark.core.utils.errorsv3 import (
    FALLBACK_EXCEPTIONS,
    APIException,
    APITimeoutError,
    InternalServiceError,
)


def _get_tool_client() -> Optional[Client]:
    client_pool = get_client_pool()
    client = client_pool.get_client("action")
    if not client:
        client = ArkActionClient()
    return client


def _ns_to_ms(ns: int) -> int:
    return int(ns / 1e6)


class ActionTool(BaseModel):
    client: ArkActionClient = Field(default_factory=_get_tool_client)
    meta_info: ArkToolMeta = Field(default_factory=ArkToolMeta)

    class Config:
        arbitrary_types_allowed = True

    async def get_schema(self) -> Dict[str, Any]:
        return {}

    @task(watch_io=False)
    async def arun(
        self,
        request: Optional[Dict[str, Any]] = None,
        dry_run: bool = False,
        **kwargs: Any,
    ) -> Union[ArkToolResponse, Tuple[ArkToolResponse, ActionDetails]]:
        created_at = _ns_to_ms(time.time_ns())
        exception: Optional[Exception] = None
        try:
            request_data = ArkToolRequest(
                action_name=self.meta_info.action_name,
                tool_name=self.meta_info.tool_name,
                parameters=request,
                dry_run=dry_run,
                timeout=(kwargs or {}).get("timeout", 60),
            )

            response = await self.client.arequest(
                api=self.__class__.__name__,
                meta_info=self.meta_info,
                request=request_data,
                response_cls=ArkToolResponse,
                extra_headers=kwargs.get("extra_headers"),
                extra_query=kwargs.get("extra_query"),
                extra_body=kwargs.get("extra_body"),
                timeout=kwargs.get("timeout"),
            )
        except Exception as e:
            if isinstance(e, asyncio.TimeoutError):
                exception = APITimeoutError(str(e))
            else:
                exception = e
        finally:
            completed_at = _ns_to_ms(time.time_ns())

        if "include_details" in kwargs and kwargs["include_details"] is True:
            action_details = ActionDetails(
                name=self.meta_info.action_name,
                count=1,
                tool_details=[
                    ToolDetails(
                        name=self.meta_info.tool_name,
                        input=request,
                        output=(
                            ToolOutput(
                                type=ToolOutputType.TOOL,
                                data=response.model_dump(
                                    exclude_unset=True, exclude_none=True
                                ),
                            )
                            if exception is None
                            else ToolOutput(
                                type=ToolOutputType.EXCEPTION,
                                data=ExceptionInfo(
                                    type=exception.__class__.__name__,
                                    message=(
                                        exception.message
                                        if isinstance(
                                            exception, (ArkAPIError, APIException)
                                        )
                                        else InternalServiceError(
                                            str(exception)
                                        ).message
                                    ),
                                ),
                            )
                        ),
                        created_at=created_at,
                        completed_at=completed_at,
                    )
                ],
            )
            if exception is not None and hasattr(exception, "message"):
                exception.message = action_details.model_dump_json(exclude_none=True)

        if exception is not None:
            if kwargs.get("special_exception_fallback") and (
                isinstance(exception, FALLBACK_EXCEPTIONS)
            ):
                response = ArkToolResponse()
            else:
                raise exception

        if "include_details" in kwargs and kwargs["include_details"] is True:
            return response, action_details
        return response

    @task()
    async def acall(
        self, request: Optional[Dict[str, Any]] = None, **kwargs: Any
    ) -> Union[ArkToolResponse, AsyncIterable[ArkToolResponse]]:
        """
        for ActionTool acall
        """
        return await self.arun(request, **kwargs)



from typing import Any

from pydantic.v1 import Field
from typing_extensions import Literal

from ark.component.llm.base import BaseChatLanguageModel, BaseChatPromptTemplate
from ark.component.prompts.custom_prompt import CustomPromptTemplate
from ark.component.prompts.rag_prompt import DEFAULT_SUMMARY_PROMPT
from ark.core.idl.maas_protocol import MaasChatResponse
from ark.core.task import task


def _default_summary_prompt() -> CustomPromptTemplate:
    return CustomPromptTemplate(
        template=DEFAULT_SUMMARY_PROMPT,
        keep_history_systems=True,
        keep_history_questions=True,
    )


class RagSummary(BaseChatLanguageModel):
    name: Literal["RagSummary"] = "RagSummary"
    template: CustomPromptTemplate = Field(default_factory=_default_summary_prompt)

    def _get_prompt_template(self) -> BaseChatPromptTemplate:
        return self.template

    @task()
    async def arun(self, *args: Any, **kwargs: Any) -> MaasChatResponse:
        return await super().arun(*args, **kwargs)

    @task()
    def run(self, *args: Any, **kwargs: Any) -> MaasChatResponse:
        return super().run(*args, **kwargs)
